
__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os

import radical.utils as ru

from ... import utils as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's spawner types
EXECUTING_NAME_POPEN   = "POPEN"
EXECUTING_NAME_SHELL   = "SHELL"
EXECUTING_NAME_SHELLFS = "SHELLFS"
EXECUTING_NAME_FLUX    = "FLUX"
EXECUTING_NAME_SLEEP   = "SLEEP"
EXECUTING_NAME_FUNCS   = "FUNCS"

# archived
#
# EXECUTING_NAME_ABDS    = "ABDS"
# EXECUTING_NAME_ORTE    = "ORTE"


# ------------------------------------------------------------------------------
#
class AgentExecutingComponent(rpu.Component):
    """
    Manage the creation of CU processes, and watch them until they are completed
    (one way or the other).  The spawner thus moves the unit from
    PendingExecution to Executing, and then to a final state (or PendingStageOut
    of course).
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id(cfg['owner'] + '.executing.%(counter)s',
                                   ru.ID_CUSTOM)

        rpu.Component.__init__(self, cfg, session)

        # if so configured, let the CU know what to use as tmp dir
        self._cu_tmp = cfg.get('cu_tmp', os.environ.get('TMP', '/tmp'))


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg, session):

        name = cfg['spawner']

        # Make sure that we are the base-class!
        if cls != AgentExecutingComponent:
            raise TypeError("Factory only available to base class!")

        from .popen    import Popen
        from .shell    import Shell
        from .shell_fs import ShellFS
        from .flux     import Flux
        from .funcs    import FUNCS
        from .sleep    import Sleep

      # from .abds     import ABDS
      # from .orte     import ORTE

        try:
            impl = {
                EXECUTING_NAME_POPEN  : Popen,
                EXECUTING_NAME_SHELL  : Shell,
                EXECUTING_NAME_SHELLFS: ShellFS,
                EXECUTING_NAME_FLUX   : Flux,
                EXECUTING_NAME_SLEEP  : Sleep,
                EXECUTING_NAME_FUNCS  : FUNCS,
              # EXECUTING_NAME_ABDS   : ABDS,
              # EXECUTING_NAME_ORTE   : ORTE,
            }[name]
            return impl(cfg, session)

        except KeyError as e:
            raise RuntimeError("AgentExecutingComponent '%s' unknown" % name) \
                from e


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os

import radical.pilot as rp


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    session = rp.Session()

    try:
        pmgr    = rp.PilotManager(session=session)
        pd_init = {'resource'      : 'local.localhost_flux',
                   'runtime'       : 60,
                   'exit_on_error' : True,
                   'cores'         : 16,
                 # 'prepare_env'   : {'env_1' : {'type'   : 'virtualenv', 
                 #                               'version': 'python3.8',
                 #                               'setup'  : ['numpy']}}
                  }
        pdesc  = rp.ComputePilotDescription(pd_init)
        pilots = pmgr.submit_pilots([pdesc] * 1)
        umgr   = rp.UnitManager(session=session)
        umgr.add_pilots(pilots)

      # for pilot in pilots:
      #     pilot.prepare_env({'env_2' : {'type'   : 'virtualenv', 
      #                                   'version': 'python3.8',
      #                                   'setup'  : ['numpy']}})
        n    = 10
        cuds = list()
        for i in range(0, n):

            cud = rp.ComputeUnitDescription()
            cud.executable       = '%s/examples/hello_rp.sh' % os.getcwd()
            cud.arguments        = ['1']
            cud.cpu_processes    = 2
            cud.cpu_threads      = 3
            cud.gpu_processes    = 0
         #  cud.named_env        = 'env_1'
            cuds.append(cud)

        umgr.submit_units(cuds)
        umgr.wait_units()

    finally:
        session.close(download=True)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import time

import radical.utils as ru
import radical.pilot as rp

glyphs = {
          rp.NEW                          : '#',
          rp.UMGR_SCHEDULING_PENDING      : '.',
          rp.UMGR_SCHEDULING              : '.',
          rp.UMGR_STAGING_INPUT_PENDING   : '.',
          rp.UMGR_STAGING_INPUT           : '.',
          rp.AGENT_STAGING_INPUT_PENDING  : '_',
          rp.AGENT_STAGING_INPUT          : '_',
          rp.AGENT_SCHEDULING_PENDING     : '_',
          rp.AGENT_SCHEDULING             : '_',
          rp.AGENT_EXECUTING_PENDING      : '@',
          rp.AGENT_EXECUTING              : '@',
          rp.AGENT_STAGING_OUTPUT_PENDING : '=',
          rp.AGENT_STAGING_OUTPUT         : '=',
          rp.UMGR_STAGING_OUTPUT_PENDING  : '=',
          rp.UMGR_STAGING_OUTPUT          : '=',
          rp.DONE                         : '+',
          rp.FAILED                       : '-',
          rp.CANCELED                     : 'x'
         }


def listen(topic, msgs):

    try:
        for msg in ru.as_list(msgs):
            if msg['cmd'] == 'update':
                for thing in msg['arg']:

                    uid   = thing['uid']
                    state = thing['state']

                  # print('%-20s: %s' % (uid, state))

                    if 'pilot' in uid:
                        cores = thing.get('description', {'cores':'?'}) \
                                     .get('cores')
                        print('\n%s [%s]: %s' % (uid, cores, state))

                    elif 'unit' in uid:
                        print(glyphs[state], end='')
    except Exception as e:
        print(e)


cfg = ru.read_json('%s/state_pubsub.cfg' % sys.argv[1])
sub = ru.zmq.Subscriber('state_pubsub', url=cfg['sub'])
sub.subscribe(topic='state_pubsub', cb=listen)

time.sleep(100000)





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from ... import utils     as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's agent staging output types
RP_ASI_NAME_DEFAULT = "DEFAULT"


# ------------------------------------------------------------------------------
#
class AgentStagingOutputComponent(rpu.Component):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id(cfg['owner'] + '.staging.output.%(counter)s',
                                   ru.ID_CUSTOM)

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Stager
    #
    @classmethod
    def create(cls, cfg, session):

        name = cfg.get('agent_staging_output_component', RP_ASI_NAME_DEFAULT)

        # Make sure that we are the base-class!
        if cls != AgentStagingOutputComponent:
            raise TypeError("Factory only available to base class!")

        from .default import Default

        try:
            impl = {
                RP_ASI_NAME_DEFAULT: Default
            }[name]

            impl = impl(cfg, session)
            return impl

        except KeyError:
            raise ValueError("AgentStagingOutputComponent '%s' defunct" % name)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import sys
import time

import radical.pilot as rp
import radical.pilot.task_overlay as rpt


# ------------------------------------------------------------------------------
#
class MyWorker(rpt.Worker):
    '''
    This class provides the required functionality to execute work requests.
    In this simple example, the worker only implements a single call: `hello`.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg):

        rp.task_overlay.Worker.__init__(self, cfg)

        self.register_call('hello', self.hello)


    # --------------------------------------------------------------------------
    #
    def hello(self, count, uid):
        '''
        important work
        '''

        self._prof.prof('dock_start', uid=uid)

        out = 'hello %5d @ %.2f [%s]' % (count, time.time(), self._uid)
      # time.sleep(0.1)

        self._prof.prof('dock_io_start', uid=uid)
        self._log.debug(out)
        self._prof.prof('dock_io_stop', uid=uid)

        self._prof.prof('dock_stop', uid=uid)
        return out


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # the `info` dict is passed to the worker as config file.
    # Create the worker class and run it's work loop.
    worker = MyWorker(sys.argv[1])
    worker.run()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__ = "MIT"


import os
import pprint
import inspect
import logging

import radical.utils as ru

from ...   import constants as rpc
from ...   import states    as rps
from .base import AgentSchedulingComponent


# ------------------------------------------------------------------------------
#
# This is an extension of the continuous scheduler with awareness of the
# file-storage capabilities on a node. The continuous data aware scheduler will
# use two data fields: availability and requirement.

# General idea:
# The availability will be obtained from the rm_node_list and assigned to
# the node list of the class. The requirement will be obtained from the cud in
# the alloc_nompi and alloc_mpi methods. Using the availability and
# requirement, the _find_resources method will return the core and gpu ids.
#
# Expected DS of the nodelist
# self.nodes = [{
#                   'name'    : 'node_1',
#                   'uid'     : xxxx,
#                   'sockets' : [  {'cores': [], 'gpus': []},
#                                  {'cores': [], 'gpus': []}]
#                   'lfs'     : 128
#               },
#               {
#                   'name'    : 'node_2',
#                   'uid'     : yyyy,
#                   'sockets' : [  {'cores': [], 'gpus': []},
#                                  {'cores': [], 'gpus': []}]
#                   'lfs'     : 256
#                },
#               ]
# Q: How should the nodes be selected for MPI based units?
# lfs : in mb

import cProfile
cprof = cProfile.Profile()


def cprof_it(func):
    def wrapper(*args, **kwargs):
        retval = cprof.runcall(func, *args, **kwargs)
        return retval
    return wrapper


def dec_all_methods(dec):
    def dectheclass(cls):
        if ru.is_main_thread():
            cprof_env = os.getenv("RADICAL_PILOT_CPROFILE_COMPONENTS", "")
            cprof_elems = cprof_env.split()
            if "CONTINUOUS" in cprof_elems:
                for name, m in inspect.getmembers(cls, inspect.ismethod):
                    setattr(cls, name, dec(m))
        return cls
    return dectheclass


# ------------------------------------------------------------------------------
#
@dec_all_methods(cprof_it)
class ContinuousSummit(AgentSchedulingComponent):
    '''
    The Continuous scheduler attempts to place threads and processes of
    a compute units onto consecutive cores, gpus and nodes in the cluster.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.nodes = None
        self._tag_history = dict()

        AgentSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # Once the component process is spawned, `initialize()` will be called
    # before control is given to the component's main loop.
    #
    def initialize(self):

        # register unit input channels
        self.register_input(rps.AGENT_SCHEDULING_PENDING,
                            rpc.AGENT_SCHEDULING_QUEUE, self._schedule_units)

        # register unit output channels
        self.register_output(rps.AGENT_EXECUTING_PENDING,
                             rpc.AGENT_EXECUTING_QUEUE)

        # we need unschedule updates to learn about units for which to free the
        # allocated cores.  Those updates MUST be issued after execution, ie.
        # by the AgentExecutionComponent.
        self.register_subscriber(rpc.AGENT_UNSCHEDULE_PUBSUB, self.unschedule_cb)

        # we don't want the unschedule above to compete with actual
        # scheduling attempts, so we move the re-scheduling of units from the
        # wait pool into a separate thread (ie. register a separate callback).
        # This is triggered by the unscheduled_cb.
        #
        # NOTE: we could use a local queue here.  Using a zmq bridge goes toward
        #       an distributed scheduler, and is also easier to implement right
        #       now, since `Component` provides the right mechanisms...
        self.register_publisher(rpc.AGENT_SCHEDULE_PUBSUB)
        self.register_subscriber(rpc.AGENT_SCHEDULE_PUBSUB, self.schedule_cb)

        # The scheduler needs the ResourceManager information which have been collected
        # during agent startup.  We dig them out of the config at this point.
        #
        # NOTE: this information is insufficient for the torus scheduler!
        self._pid                   = self._cfg['pid']
        self._rm_info             = self._cfg['rm_info']
        self._rm_lm_info          = self._cfg['rm_info']['lm_info']
        self._rm_node_list        = self._cfg['rm_info']['node_list']
        self._rm_sockets_per_node = self._cfg['rm_info']['sockets_per_node']
        self._rm_cores_per_socket = self._cfg['rm_info']['cores_per_socket']
        self._rm_gpus_per_socket  = self._cfg['rm_info']['gpus_per_socket']
        self._rm_lfs_per_node     = self._cfg['rm_info']['lfs_per_node']

        if not self._rm_node_list:
            raise RuntimeError("ResourceManager %s didn't _configure node_list."
                              % self._rm_info['name'])

        if self._rm_cores_per_socket is None:
            raise RuntimeError("ResourceManager %s didn't _configure cores_per_socket."
                              % self._rm_info['name'])

        if self._rm_sockets_per_node is None:
            raise RuntimeError("ResourceManager %s didn't _configure sockets_per_node."
                              % self._rm_info['name'])

        if self._rm_gpus_per_socket is None:
            raise RuntimeError("ResourceManager %s didn't _configure gpus_per_socket."
                              % self._rm_info['name'])

        # create and initialize the wait pool
        self._wait_pool = list()      # pool of waiting units
        self._wait_lock = ru.RLock()  # look on the above pool
        self._slot_lock = ru.RLock()  # lock slot allocation/deallocation

        # configure the scheduler instance
        self._configure()
        self._log.debug("slot status after  init      : %s",
                        self.slot_status())


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # * oversubscribe:
        #   Cray's aprun for example does not allow us to oversubscribe CPU
        #   cores on a node, so we can't, say, run n CPU processes on an n-core
        #   node, and than add one additional process for a GPU application.
        # If oversubscribe` is set to False (which is the default for now),
        #   we'll prevent that behavior by allocating one additional CPU core
        #   for each set of requested GPU processes.
        #   FIXME: I think our scheme finds the wrong core IDs for GPU process
        #          startup - i.e. not the reserved ones.
        self._oversubscribe = self._cfg.get('oversubscribe', True)

        # * scattered:
        #   This is the continuous scheduler, because it attempts to allocate
        #   a *continuous* set of cores/nodes for a unit.  It does, however,
        #   also allow to scatter the allocation over discontinuous nodes if
        #   this option is set.  This implementation is not optimized for the
        #   scattered mode!  The default is 'False'.
        #
        self._scattered = self._cfg.get('scattered', False)


        # * cross_socket_threads:
        # We now capture the knowledge of sockets on nodes and the relationship
        # of cores to sockets and gpus to sockets. We want to ensure that cpu
        # threads belonging to a single process do not cross sockets if the
        # system does not allow it. The default is True, based on discussions
        # in the Summit meeting, and will be set to False for Summit in the
        # resource config.
        self._cross_socket_threads = self._cfg.get('cross_socket_threads', True)

        # NOTE:  for non-oversubscribing mode, we reserve a number of cores
        #        for the GPU processes - even if those GPUs are not used by
        #        a specific workload.  In this case we rewrite the node list and
        #        substract the respective number of available cores per socket.
        if not self._oversubscribe:

            if self._rm_cores_per_socket <= self._rm_gpus_per_socket:
                raise RuntimeError('oversubscription mode requires more cores')

            self._rm_cores_per_socket -= self._rm_gpus_per_socket


        # since we just changed this fundamental setting, we need to
        # recreate the nodelist.
        self.nodes = []
        for node, node_uid in self._rm_node_list:

            node_entry = {'name'   : node,
                          'uid'    : node_uid,
                          'sockets': list(),
                          'lfs'    : self._rm_lfs_per_node}

            for socket in range(self._rm_sockets_per_node):
                node_entry['sockets'].append({
                    'cores': [rpc.FREE] * self._rm_cores_per_socket,
                    'gpus' : [rpc.FREE] * self._rm_gpus_per_socket
                })

            self.nodes.append(node_entry)


    # --------------------------------------------------------------------------
    #
    # Change the reserved state of slots (rpc.FREE or rpc.BUSY)
    #
    # NOTE: any scheduler implementation which uses a different nodelist
    #       structure MUST overload this method.
    #
    def _change_slot_states(self, slots, new_state):
        '''
        This function is used to update the state for a list of slots that
        have been allocated or deallocated.  For details on the data structure,
        see top of `base.py`.
        '''
        # This method needs to change if the DS changes.
        # Current slot DS:
        # slots = {'nodes': [{'name'    : node_name,
        #                     'uid'     : node_uid,
        #                     'core_map': core_map,
        #                     'gpu_map' : gpu_map,
        #                     'lfs'     : {'size': lfs,
        #                                  'path': self._rm_lfs_per_node['path']
        #                                 }
        #                    }],
        #          'cores_per_node' : cores_per_node,
        #          'gpus_per_node'  : gpus_per_node,
        #          'lfs_per_node'   : self._rm_lfs_per_node,
        #          'lm_info'        : self._rm_lm_info
        #          }
        #
        # self.nodes = [{
        #                   'name'    : 'node_1',
        #                   'uid'     : xxxx,
        #                   'sockets' : [   {'cores': [], 'gpus': []},
        #                                   {'cores': [], 'gpus': []}]
        #                   'lfs'     : 128}],


        # for node_name, node_uid, cores, gpus in slots['nodes']:
        for node in slots['nodes']:

            # Find the entry in the the slots list

            # TODO: [Optimization] Assuming 'uid' is the ID of the node, it
            #       seems a bit wasteful to have to look at all of the nodes
            #       available for use if at most one node can have that uid.
            #       Maybe it would be worthwhile to simply keep a list of nodes
            #       that we would read, and keep a dictionary that maps the uid
            #       of the node to the location on the list?

            for entry in self.nodes:
                if entry['uid'] == node['uid']:
                    break

            assert(entry), 'missing node %s' % node['uid']

            # STRONG ASSUMPTION!!!
            # We assume that the core and gpu ids are continuous on a socket
            # indexed with core_id=0 on socket_id=0 and gpu_id= on socket_id=0.
            # For example, with the following node config for Summit:
            # sockets_per_node=2, cores_per_socket=21, gpus_per_socket=3
            # Our assumption leads to the following id relations:

            # node['sockets'] = [
            #     {   'uid': 0,     # not actually present in socket list
            #         'core_ids': [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
            #                      10,11,12,13,14,15,16,17,18,19,
            #                      20],
            #         'gpu_ids':  [0,1,2]
            #     }
            #     {   'uid': 1,    # not actually present in socket list
            #         'core_ids': [21,22,23,24,25,26,27,28,29,30,
            #                      31,32,33,34,35,36,37,38,39,40,
            #                      41],
            #         'gpu_ids':  [3,4,5]
            #     }
            # ]

            # iterate over resources in the slot, and update state
            for cslot in node['core_map']:
                for core in cslot:
                    socket  = core / self._rm_cores_per_socket
                    core_id = core % self._rm_cores_per_socket
                    entry['sockets'][socket]['cores'][core_id] = new_state

            for gslot in node['gpu_map']:
                for gpu in gslot:
                    socket = gpu / self._rm_gpus_per_socket
                    gpu_id = gpu % self._rm_gpus_per_socket
                    entry['sockets'][socket]['gpus'][gpu_id] = new_state

            if entry['lfs']['path']:
                if new_state == rpc.BUSY:
                    entry['lfs']['size'] -= node['lfs']['size']
                else:
                    entry['lfs']['size'] += node['lfs']['size']


    # --------------------------------------------------------------------------
    #
    # Overloaded from Base
    def slot_status(self):
        '''
        Returns a multi-line string corresponding to the status of the node list
        '''

        ret = "|"
        for node in self.nodes:
            for socket in node['sockets']:
                for core in socket['cores']:
                    if core == rpc.FREE:
                        ret += '-'
                    else:
                        ret += '#'
                ret += ':'
                for gpu in socket['gpus']:
                    if gpu == rpc.FREE:
                        ret += '-'
                    else:
                        ret += '#'
                ret += '|'

        return ret


    # --------------------------------------------------------------------------
    #
    def _try_allocation(self, unit):
        """
        attempt to allocate cores/gpus for a specific unit.
        """

        uid = unit['uid']

        # needs to be locked as we try to acquire slots here, but slots are
        # freed in a different thread.  But we keep the lock duration short...
        with self._slot_lock:

            self._prof.prof('schedule_try', uid=uid)
            unit['slots'] = self._allocate_slot(unit)


        # the lock is freed here
        if not unit['slots']:

            # signal the unit remains unhandled (Fales signals that failure)
            self._prof.prof('schedule_fail', uid=uid)
            return False


        node_uids = [node['uid'] for node in unit['slots']['nodes']]
        self._tag_history[uid] = node_uids

        # got an allocation, we can go off and launch the process
        self._prof.prof('schedule_ok', uid=uid)

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug("after  allocate   %s: %s", uid,
                            self.slot_status())
            self._log.debug("%s [%s/%s] : %s", uid,
                            unit['description']['cpu_processes'],
                            unit['description']['gpu_processes'],
                            pprint.pformat(unit['slots']))

        # True signals success
        return True


    # --------------------------------------------------------------------------
    #
    def _allocate_slot(self, unit):
        '''
        This is the main method of this implementation, and is triggered when
        a unit needs to be mapped to a set of cores / gpus.  We make
        a distinction between MPI and non-MPI units (non-MPI processes MUST be
        on the same node).
        '''

        uid = unit['uid']
        cud = unit['description']

        # single_node allocation is enforced for non-message passing tasks
        if cud['cpu_process_type'] in [rpc.MPI] or \
           cud['gpu_process_type'] in [rpc.MPI]:
            slots = self._alloc_mpi(unit)
        else:
            slots = self._alloc_nompi(unit)

        if slots:
            # the unit was placed, we need to reflect the allocation in the
            # nodelist state (BUSY)
            self._change_slot_states(slots, rpc.BUSY)

        return slots


    # --------------------------------------------------------------------------
    #
    def _release_slot(self, slots):
        '''
        This method is called when previously aquired resources are not needed
        anymore.  `slots` are the resource slots as previously returned by
        `_allocate_slots()`.
        '''

        # reflect the request in the nodelist state (set to `FREE`)
        self._change_slot_states(slots, rpc.FREE)

    # --------------------------------------------------------------------------
    #
    def _find_resources(self, node, requested_cores, requested_gpus,
                        requested_lfs, core_chunk=1, partial=False,
                        lfs_chunk=1, gpu_chunk=1):
        '''
        Find up to the requested number of free cores and gpus in the node.
        This call will return two lists, for each matched set.  If the core
        does not have sufficient free resources to fulfill *both* requests, two
        empty lists are returned.  The call will *not* change the allocation
        status of the node, atomicity must be guaranteed by the caller.

        We don't care about continuity within a single node - cores `[1,5]` are
        assumed to be as close together as cores `[1,2]`.

        When `chunk` is set, only sets of exactly that size (or multiples
        thereof) are considered valid allocations.  The use case is OpenMP
        support, where each process is expected to create a certain number of
        threads, which thus must have slots on the same node available.

        NOTE: chunking is only applied to cores at this point.

        When `partial` is set to `True`, this method is allowed to return
        a *partial* match, so to find less cores, gpus, and local_fs then
        requested (but the call will never return more than requested).
        '''

        # list of core and gpu ids available in this node.
        cores = list()
        gpus  = list()
        lfs   = 0

        # first count the number of free cores, gpus, and local file storage.
        # This is way quicker than actually finding the core IDs.
        free_cores = 0
        free_gpus  = 0
        free_lfs   = node['lfs']['size']
        free_cores_per_socket = list()
        free_gpus_per_socket  = list()

        for socket in node['sockets']:
            free_cores += socket['cores'].count(rpc.FREE)
            free_gpus  += socket['gpus'].count(rpc.FREE)

            free_cores_per_socket.append(socket['cores'].count(rpc.FREE))
            free_gpus_per_socket.append(socket['gpus'].count(rpc.FREE))


        alloc_lfs   = 0
        alloc_cores = 0
        alloc_gpus  = 0

        if partial:
            # For partial requests the check simplifies: we just check if we
            # have either, some cores *or* gpus *or* local_fs, to serve the
            # request
            if (requested_cores and not free_cores) and \
               (requested_gpus  and not free_gpus)  and \
               (requested_lfs   and not free_lfs):
                return [], [], None

            if requested_lfs and \
                ((requested_cores and not free_cores) and
                 (requested_gpus  and not free_gpus)):
                return [], [], None

        else:
            # For non-partial requests (ie. full requests): its a no-match if
            # either the cpu or gpu request cannot be served.
            if  requested_cores > free_cores or \
                requested_gpus  > free_gpus  or \
                requested_lfs   > free_lfs   :
                return [], [], None

        # We can serve the partial or full request - alloc the chunks we need
        # FIXME: chunk gpus, too?
        # We need to land enough procs on a node such that the cores,
        # lfs and gpus requested per proc is available on the same node

        num_procs = dict()

        if requested_lfs:
            alloc_lfs = min(requested_lfs, free_lfs)
            num_procs['lfs'] = alloc_lfs / lfs_chunk

        if requested_cores:
            self._log.debug('req cores: %s of %s [%s]', requested_cores,
                    free_cores, core_chunk)

            if self._cross_socket_threads:
                alloc_cores = min(requested_cores, free_cores)
                num_procs['cores'] = alloc_cores / core_chunk

            else:
                # If no cross socket threads are allowed, we first make sure that the total
                # number of cores is greater than the core_chunk. If not, we can't use the
                # current node.
                if sum(free_cores_per_socket) < core_chunk:
                    return [], [], None

                # If we can use this node, then we visit each socket of the node and
                # determine the number of continuous core_chunks that can be use on the
                # socket till we either run out of continuous cores or have acquired
                # the requested number of cores.
                usable_cores = 0

                # Determine maximum procs on each socket - required during assignment
                max_procs_on_socket = [0 for _ in range(self._rm_sockets_per_node)]
                self._log.debug('mpos %s', max_procs_on_socket)

                for socket_id, free_cores in enumerate(free_cores_per_socket):
                    tmp_num_cores = free_cores

                    while tmp_num_cores >= core_chunk and \
                          usable_cores  <  requested_cores:

                        usable_cores  += core_chunk
                        tmp_num_cores -= core_chunk
                        max_procs_on_socket[socket_id] += 1

                    if usable_cores == requested_cores:
                        break

                # We convert the number of usable cores into a number of procs so that
                # we can find the minimum number of procs across lfs, cpus, gpus that can
                # be allocated on this node
                num_procs['cores'] = usable_cores / core_chunk

        if requested_gpus:
            alloc_gpus = min(requested_gpus, free_gpus)
            num_procs['gpus'] = alloc_gpus / gpu_chunk

        # Find normalized lfs, cores and gpus
        if requested_cores: alloc_cores = num_procs['cores'] * core_chunk
        if requested_gpus : alloc_gpus  = num_procs['gpus']  * gpu_chunk
        if requested_lfs  : alloc_lfs   = num_procs['lfs']   * lfs_chunk
        self._log.debug('alc : %s %s %s', alloc_cores, alloc_gpus, alloc_lfs)

        # Maximum number of processes allocatable on a socket
        if self._cross_socket_threads:
            max_procs_on_socket = [num_procs['cores']
                                   for _ in range(self._rm_sockets_per_node)]
            self._log.debug('max_procs_on_socket %s', max_procs_on_socket)

        # now dig out the core IDs.
        for socket_idx, socket in enumerate(node['sockets']):
            procs_on_socket = 0
            for core_idx, state in enumerate(socket['cores']):

                # break if we have enough cores, else continue to pick FREE ones
                if alloc_cores == len(cores):
                    break
                if state == rpc.FREE:
                    cores.append(socket_idx * self._rm_cores_per_socket
                                            + core_idx)

                # check if we have placed one complete process on current socket
                if cores and (len(cores) % core_chunk == 0):
                    procs_on_socket += 1

                # if we have reached max procs on socket, move to next socket
                if procs_on_socket == max_procs_on_socket[socket_idx]:
                    break

            # break if we have enough cores, else continue to pick FREE ones
            if alloc_cores == len(cores):
                break

        # now dig out the GPU IDs.
        for socket_idx, socket in enumerate(node['sockets']):
            for gpu_idx, state in enumerate(socket['gpus']):

                # break if we have enough gpus, else continue to pick FREE ones
                if alloc_gpus == len(gpus):
                    break
                if state == rpc.FREE:
                    gpus.append(socket_idx*self._rm_gpus_per_socket+ gpu_idx)

            # break if we have enough gpus, else continue to pick FREE ones
            if alloc_gpus == len(gpus):
                break

        return cores, gpus, alloc_lfs


    # --------------------------------------------------------------------------
    #
    def _get_node_maps(self, cores, gpus, threads_per_proc):
        """
        For a given set of cores and gpus, chunk them into sub-sets so that
        each sub-set can host one application process and all threads of that
        process.  Note that we currently consider all GPU applications to be
        single-threaded.
        For more details, see top level comment of `base.py`.
        """

        core_map = list()
        gpu_map  = list()

        # make sure the core sets can host the requested number of threads
        assert(not len(cores) % threads_per_proc)
        n_procs =  len(cores) / threads_per_proc

        idx = 0
        for p in range(n_procs):
            p_map = list()
            for t in range(threads_per_proc):
                p_map.append(cores[idx])
                idx += 1
            core_map.append(p_map)

        if idx != len(cores):
            self._log.debug('%s -- %s -- %s -- %s',
                            idx, len(cores), cores, n_procs)
        assert(idx == len(cores))

        # gpu procs are considered single threaded right now (FIXME)
        for g in gpus:
            gpu_map.append([g])

        return core_map, gpu_map


    # --------------------------------------------------------------------------
    #
    def _alloc_nompi(self, unit):
        """
        Find a suitable set of cores and gpus *within a single node*.

        Input:
        cud: Compute Unit description. Needs to specify at least one CPU
        process and one thread per CPU process, or one GPU process.
        """

        uid = unit['uid']
        cud = unit['description']

        # dig out the allocation request details
        requested_procs  = cud['cpu_processes']
        threads_per_proc = cud['cpu_threads']
        requested_gpus   = cud['gpu_processes']
        requested_lfs    = cud['lfs_per_process']
        lfs_chunk        = requested_lfs if requested_lfs > 0 else 1

        # make sure that processes are at least single-threaded
        if not threads_per_proc:
            threads_per_proc = 1

        # cores needed for all threads and processes
        requested_cores = requested_procs * threads_per_proc

        if not self._cross_socket_threads:
            if threads_per_proc > self._rm_cores_per_socket:
                raise ValueError('cu does not fit on socket')

        cores_per_node = self._rm_cores_per_socket * self._rm_sockets_per_node
        gpus_per_node  = self._rm_gpus_per_socket  * self._rm_sockets_per_node
        lfs_per_node   = self._rm_lfs_per_node['size']

        # make sure that the requested allocation fits within the resources
        if  requested_cores > cores_per_node or \
            requested_gpus  > gpus_per_node  or \
            requested_lfs   > lfs_per_node   :

            txt  = 'Non-mpi unit %s does not fit onto node \n' % uid
            txt += '   cores: %s >? %s \n' % (requested_cores, cores_per_node)
            txt += '   gpus : %s >? %s \n' % (requested_gpus,  gpus_per_node)
            txt += '   lfs  : %s >? %s'    % (requested_lfs,   lfs_per_node)
            raise ValueError(txt)

        # ok, we can go ahead and try to find a matching node
        cores     = list()
        gpus      = list()
        lfs       = None
        node_name = None
        node_uid  = None
        tag       = cud.get('tag')

        for node in self.nodes:  # FIXME optimization: iteration start

            # If unit has a tag, check if the tag is in the tag_history dict,
            # else it is a invalid tag, continue as if the unit does not have
            # a tag
            # If the unit has a valid tag, find the node that matches the
            # tag from tag_history dict
            if tag and tag in self._tag_history:
                if node['uid'] not in self._tag_history[tag]:
                    continue

            # attempt to find the required number of cores and gpus on this
            # node - do not allow partial matches.
            cores, gpus, lfs = self._find_resources(node=node,
                                                    requested_cores=requested_cores,
                                                    requested_gpus=requested_gpus,
                                                    requested_lfs=requested_lfs,
                                                    partial=False,
                                                    lfs_chunk=lfs_chunk,
                                                    core_chunk = threads_per_proc
                                                    )
            if  len(cores) == requested_cores and \
                len(gpus)  == requested_gpus:

                # we found the needed resources - break out of search loop
                node_uid  = node['uid']
                node_name = node['name']
                break

        # If we did not find any node to host this request, return `None`
        if not node_name:
            return None

        # We have to communicate to the launcher where exactly processes are to
        # be placed, and what cores are reserved for application threads.  See
        # the top level comment of `base.py` for details on the data structure
        # used to specify process and thread to core mapping.
        core_map, gpu_map = self._get_node_maps(cores, gpus, threads_per_proc)

        # We need to specify the node lfs path that the unit needs to use.
        # We set it as an environment variable that gets loaded with cud
        # executable.
        # Assumption enforced: The LFS path is the same across all nodes.
        cud['environment']['NODE_LFS_PATH'] = self._rm_lfs_per_node['path']

        # all the information for placing the unit is acquired - return them
        slots = {'nodes': [{'name'    : node_name,
                            'uid'     : node_uid,
                            'core_map': core_map,
                            'gpu_map' : gpu_map,
                            'lfs'     : {'size': lfs,
                                         'path': self._rm_lfs_per_node['path']
                                        }
                           }],
                 'cores_per_node': cores_per_node,
                 'gpus_per_node' : gpus_per_node,
                 'lfs_per_node'  : lfs_per_node,
                 'lm_info'       : self._rm_lm_info
                 }

        return slots


    # --------------------------------------------------------------------------
    #
    #
    def _alloc_mpi(self, unit):
        """
        Find an available set of slots, potentially across node boundaries.  By
        default, we only allow for partial allocations on the first and last
        node - but all intermediate nodes MUST be completely used (this is the
        'CONTINUOUS' scheduler after all).

        If the scheduler is configured with `scattered=True`, then that
        constraint is relaxed, and any set of slots (be it continuous across
        nodes or not) is accepted as valid allocation.

        No matter the mode, we always make sure that we allocate in chunks of
        'threads_per_proc', as otherwise the application would not be able to
        spawn the requested number of threads on the respective node.
        """

        uid = unit['uid']
        cud = unit['description']

        # dig out the allocation request details
        requested_procs  = cud['cpu_processes']
        threads_per_proc = cud['cpu_threads']
        requested_gpus   = cud['gpu_processes']
        requested_lfs_per_process = cud['lfs_per_process']

        # make sure that processes are at least single-threaded
        if not threads_per_proc:
            threads_per_proc = 1

        # cores needed for all threads and processes
        requested_cores = requested_procs * threads_per_proc

        # We allocate the same lfs per process (agreement)
        requested_lfs = requested_lfs_per_process * requested_procs

        # First and last nodes can be a partial allocation - all other nodes
        # can only be partial when `scattered` is set.
        #
        # Iterate over all nodes until we find something. Check if it fits the
        # allocation mode and sequence.  If not, start over with the next node.
        # If it matches, add the slots found and continue to next node.
        #
        # FIXME: persistent node index
        #
        # Things are complicated by chunking: we only accept chunks of
        # 'threads_per_proc', as otherwise threads would need to be distributed
        # over nodes, which is not possible for the multi-system-image clusters
        # this scheduler assumes.
        #
        #   - requested_cores > cores_per_node
        #   - cores_per_node  % threads_per_proc != 0
        #   - scattered is False
        #
        # but it can fail for less cores, too, if the partial first and last
        # allocation are not favorable.  We thus raise an exception for
        # requested_cores > cores_per_node on impossible full-node-chunking

        cores_per_node = self._rm_cores_per_socket * self._rm_sockets_per_node
        gpus_per_node  = self._rm_gpus_per_socket  * self._rm_sockets_per_node
        lfs_per_node   = self._rm_lfs_per_node

        if not self._cross_socket_threads:
            if threads_per_proc > self._rm_cores_per_socket:
                raise ValueError('Number of threads greater than that available on a socket')

        # we always fail when too many threads are requested
        if threads_per_proc > cores_per_node:
            raise ValueError('too many threads requested')

        if requested_lfs_per_process > lfs_per_node['size']:
            raise ValueError('Not enough LFS for the MPI-process')


        # set conditions to find the first matching node
        is_first      = True
        is_last       = False
        alloced_cores = 0
        alloced_gpus  = 0
        alloced_lfs   = 0

        slots = {'nodes': list(),
                 'cores_per_node': cores_per_node,
                 'gpus_per_node' : gpus_per_node,
                 'lfs_per_node'  : lfs_per_node,
                 'lm_info'       : self._rm_lm_info,
                }

        tag = cud.get('tag')

        # start the search
        for node in self.nodes:

            node_uid  = node['uid']
            node_name = node['name']

            # If unit has a tag, check if the tag is in the tag_history dict,
            # else it is a invalid tag, continue as if the unit does not have
            # a tag
            # If the unit has a valid tag, find the node that matches the
            # tag from tag_history dict
            if tag and tag in self._tag_history:
                if node['uid'] not in self._tag_history[tag]:
                    continue

            # if only a small set of cores/gpus remains unallocated (ie. less
            # than node size), we are in fact looking for the last node.  Note
            # that this can also be the first node, for small units.
            if  requested_cores - alloced_cores <= cores_per_node and \
                requested_gpus  - alloced_gpus  <= gpus_per_node  and \
                requested_lfs   - alloced_lfs   <= lfs_per_node['size']:
                is_last = True

            # we allow partial nodes on the first and last node, and on any
            # node if a 'scattered' allocation is requested.
            if is_first or self._scattered or is_last:
                partial = True
            else:
                partial = False

            # now we know how many cores/gpus we still need at this point - but
            # we only search up to node-size on this node.  Duh!
            find_cores = min(requested_cores - alloced_cores, cores_per_node)
            find_gpus  = min(requested_gpus  - alloced_gpus,  gpus_per_node)
            find_lfs   = min(requested_lfs   - alloced_lfs,   lfs_per_node['size'])

            # under the constraints so derived, check what we find on this node
            cores, gpus, lfs = self._find_resources(node=node,
                                                    requested_cores=find_cores,
                                                    requested_gpus=find_gpus,
                                                    requested_lfs=find_lfs,
                                                    core_chunk=threads_per_proc,
                                                    partial=partial,
                                                    lfs_chunk=requested_lfs_per_process)

            # Skip nodes that provide only lfs and no cores
            if not cores and lfs:
                continue

            if not cores and not gpus and not lfs:

                # this was not a match. If we are in  'scattered' mode, we just
                # ignore this node.  Otherwise we have to restart the search.
                if not self._scattered:
                    is_first      = True
                    is_last       = False
                    alloced_cores = 0
                    alloced_gpus  = 0
                    alloced_lfs   = 0
                    slots['nodes'] = list()

                # try next node
                continue

            # we found something - add to the existing allocation, switch gears
            # (not first anymore), and try to find more if needed
            self._log.debug('found %s cores, %s gpus, %s lfs', cores, gpus, lfs)
            core_map, gpu_map = self._get_node_maps(cores, gpus, threads_per_proc)

            # We need to specify the node lfs path that the unit needs to use.
            # We set it as an environment variable that gets loaded with cud
            # executable.
            # Assumption enforced: The LFS path is the same across all nodes.
            lfs_path = self._rm_lfs_per_node['path']
            if 'NODE_LFS_PATH' not in cud['environment']:
                cud['environment']['NODE_LFS_PATH'] = lfs_path

            slots['nodes'].append({'name'    : node_name,
                                   'uid'     : node_uid,
                                   'core_map': core_map,
                                   'gpu_map' : gpu_map,
                                   'lfs'     : {'size': lfs,
                                                'path': lfs_path}})

            alloced_cores += len(cores)
            alloced_gpus  += len(gpus)
            alloced_lfs   += lfs

            is_first = False

            # or maybe don't continue the search if we have in fact enough!
            if  alloced_cores == requested_cores and \
                alloced_gpus  == requested_gpus  and \
                alloced_lfs   == requested_lfs:
                # we are done
                break

        # if we did not find enough, there is not much we can do at this point
        if  alloced_cores < requested_cores or \
            alloced_gpus  < requested_gpus  or \
            alloced_lfs   < requested_lfs:
            return None  # signal failure

        # this should be nicely filled out now - return
        return slots


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2014-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import os
import sys
import copy
import stat
import time
import pprint
import subprocess          as sp
import multiprocessing     as mp

import radical.utils       as ru

from ..   import utils     as rpu
from ..   import states    as rps
from ..   import constants as rpc
from ..db import DBSession

from .resource_manager import ResourceManager
from .launch_method    import LaunchMethod


# ------------------------------------------------------------------------------
#
class Agent_0(rpu.Worker):

    '''
    This is the main agent.  It starts sub-agents and watches them.  If any of
    the sub-agents die, it will shut down the other sub-agents and itself.

    This class inherits the rpu.Worker, so that it can use its communication
    bridges and callback mechanisms.  Specifically, it will pull the DB for
    new tasks to be exexuted and forwards them to the agent's component
    network (see `work()`).  It will also watch the DB for any commands to be
    forwarded (pilot termination, task cancelation, etc), and will take care
    of heartbeat messages to be sent to the client module.  To do all this, it
    initializes a DB connection in `initialize()`.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._cfg     = cfg
        self._pid     = cfg.pid
        self._pmgr    = cfg.pmgr
        self._pwd     = cfg.pilot_sandbox
        self._session = session
        self._log     = session._log

        self._starttime   = time.time()
        self._final_cause = None

        # this is the earliest point to sync bootstrap and agent profiles
        prof = ru.Profiler(ns='radical.pilot', name='agent.0')
        prof.prof('sync_rel', uid=cfg.pid, msg='agent.0')
        prof.prof('hostname', uid=cfg.pid, msg=ru.get_hostname())

        # connect to MongoDB for state push/pull
        self._connect_db()

        # configure ResourceManager before component startup, as components need
        # ResourceManager information for function (scheduler, executor)
        self._configure_rm()

        # ensure that app communication channels are visible to workload
        self._configure_app_comm()

        # expose heartbeat channel to sub-agents, bridges and components,
        # and start those
        self._cmgr = rpu.ComponentManager(self._cfg)
        self._cfg.heartbeat = self._cmgr.cfg.heartbeat

        self._cmgr.start_bridges()
        self._cmgr.start_components()

        # create the sub-agent configs and start the sub agents
        self._write_sa_configs()
        self._start_sub_agents()   # TODO: move to cmgr?

        # at this point the session is up and connected, and it should have
        # brought up all communication bridges and components.  We are
        # ready to rumble!
        rpu.Worker.__init__(self, self._cfg, session)

        # run our own slow-paced heartbeat monitor to watch pmgr heartbeats
        # FIXME: we need to get pmgr freq
        freq = 10
        tint = freq / 3
        tout = freq * 3
        self._hb = ru.Heartbeat(uid=self._pid,
                                timeout=tout,
                                interval=tint,
                                beat_cb=self._hb_check,  # no own heartbeat(pmgr pulls)
                                term_cb=self._hb_term_cb,
                                log=self._log)
        self._hb.start()

        # register pmgr heartbeat
        self._log.info('hb init for %s', self._pmgr)
        self._hb.beat(uid=self._pmgr)


    # --------------------------------------------------------------------------
    #
    def _hb_check(self):

        self._log.debug('hb check')


    # --------------------------------------------------------------------------
    #
    def _hb_term_cb(self, msg=None):

        self._cmgr.close()
        self._log.warn('hb termination: %s', msg)

        return None


    # --------------------------------------------------------------------------
    #
    def _connect_db(self):

        # Check for the RADICAL_PILOT_DB_HOSTPORT env var, which will hold
        # the address of the tunnelized DB endpoint. If it exists, we
        # overrule the agent config with it.
        hostport = os.environ.get('RADICAL_PILOT_DB_HOSTPORT')
        if hostport:
            host, port = hostport.split(':', 1)
            dburl      = ru.Url(self._cfg.dburl)
            dburl.host = host
            dburl.port = port
            self._cfg.dburl = str(dburl)

        self._dbs = DBSession(sid=self._cfg.sid, dburl=self._cfg.dburl,
                              cfg=self._cfg, log=self._log)

    # --------------------------------------------------------------------------
    #
    def _configure_rm(self):

        # Create ResourceManager which will give us the set of agent_nodes to
        # use for sub-agent startup.  Add the remaining ResourceManager
        # information to the config, for the benefit of the scheduler).

        self._rm = ResourceManager.create(name=self._cfg.resource_manager,
                                           cfg=self._cfg, session=self._session)

        # add the resource manager information to our own config
        self._cfg['rm_info'] = self._rm.rm_info


    # --------------------------------------------------------------------------
    #
    def _configure_app_comm(self):

        # if the pilot description contains a request for application comm
        # channels, merge those into the agent config
        #
        # FIXME: this needs to start the app_comm bridges
        app_comm = self._cfg.get('app_comm')
        if app_comm:
            if isinstance(app_comm, list):
                app_comm = {ac: {'bulk_size': 0,
                                 'stall_hwm': 1,
                                 'log_level': 'error'} for ac in app_comm}
            for ac in app_comm:
                if ac in self._cfg['bridges']:
                    raise ValueError('reserved app_comm name %s' % ac)
                self._cfg['bridges'][ac] = app_comm[ac]


        # some of the bridge addresses also need to be exposed to the workload
        if app_comm:
            if 'unit_environment' not in self._cfg:
                self._cfg['unit_environment'] = dict()
            for ac in app_comm:
                if ac not in self._cfg['bridges']:
                    raise RuntimeError('missing app_comm %s' % ac)
                self._cfg['unit_environment']['RP_%s_IN' % ac.upper()] = \
                        self._cfg['bridges'][ac]['addr_in']
                self._cfg['unit_environment']['RP_%s_OUT' % ac.upper()] = \
                        self._cfg['bridges'][ac]['addr_out']


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # registers the staging_input_queue as this is what we want to push
        # units to
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE)

        # register the command callback which pulls the DB for commands
        self.register_timed_cb(self._agent_command_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # register idle callback to pull for units
        self.register_timed_cb(self._check_units_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # sub-agents are started, components are started, bridges are up: we are
        # ready to roll!  Update pilot state.
        pilot = {'type'             : 'pilot',
                 'uid'              : self._pid,
                 'state'            : rps.PMGR_ACTIVE,
                 'resource_details' : {
                     'lm_info'      : self._rm.lm_info.get('version_info'),
                     'lm_detail'    : self._rm.lm_info.get('lm_detail'),
                     'rm_info'      : self._rm.rm_info},
                 '$set'             : ['resource_details']}
        self.advance(pilot, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def work(self):

        # all work is done in the registered callbacks
        time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def stage_output(self):

        if  os.path.isfile('./staging_output.txt'):

            if not os.path.isfile('./staging_output.tgz'):

                cmd = 'tar zcvf staging_output.tgz $(cat staging_output.txt)'
                out, err, ret = ru.sh_callout(cmd, shell=True)

                if ret:
                    self._log.debug('out: %s', out)
                    self._log.debug('err: %s', err)
                    self._log.error('output tarring failed: %s', cmd)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # tar up output staging data
        self._log.debug('stage output parent')
        self.stage_output()

        # tear things down in reverse order
        self._hb.stop()
        self._cmgr.close()

        if self._rm:
            self._rm.stop()

        if   self._final_cause == 'timeout'  : state = rps.DONE
        elif self._final_cause == 'cancel'   : state = rps.CANCELED
        elif self._final_cause == 'sys.exit' : state = rps.CANCELED
        else                                 : state = rps.FAILED

        # NOTE: we do not push the final pilot state, as that is done by the
        #       bootstrapper *after* this pilot *actually* finished.
        with open('./killme.signal', 'w') as fout:
            fout.write('%s\n' % state)

        # we don't rely on the existence / viability of the update worker at
        # that point.
        self._log.debug('update db state: %s: %s', state, self._final_cause)
        self._log.info('rusage: %s', rpu.get_rusage())

        out, err, log = '', '', ''

        try   : out   = open('./agent.0.out', 'r').read(1024)
        except: pass
        try   : err   = open('./agent.0.err', 'r').read(1024)
        except: pass
        try   : log   = open('./agent.0.log', 'r').read(1024)
        except: pass

        ret = self._dbs._c.update({'type' : 'pilot',
                                   'uid'  : self._pid},
                                  {'$set' : {'stdout' : rpu.tail(out),
                                             'stderr' : rpu.tail(err),
                                             'logfile': rpu.tail(log),
                                             'state'  : state},
                                   '$push': {'states' : state}
                                  })
        self._log.debug('update ret: %s', ret)


    # --------------------------------------------------------------------
    #
    def _write_sa_configs(self):

        # we have all information needed by the subagents -- write the
        # sub-agent config files.

        # write deep-copies of the config for each sub-agent (sans from agent.0)
        for sa in self._cfg.get('agents', {}):

            assert(sa != 'agent.0'), 'expect subagent, not agent.0'

            # use our own config sans agents/components/bridges as a basis for
            # the sub-agent config.
            tmp_cfg = copy.deepcopy(self._cfg)
            tmp_cfg['agents']     = dict()
            tmp_cfg['components'] = dict()
            tmp_cfg['bridges']    = dict()

            # merge sub_agent layout into the config
            ru.dict_merge(tmp_cfg, self._cfg['agents'][sa], ru.OVERWRITE)

            tmp_cfg['uid']   = sa
            tmp_cfg['aid']   = sa
            tmp_cfg['owner'] = 'agent.0'

            ru.write_json(tmp_cfg, './%s.cfg' % sa)


    # --------------------------------------------------------------------------
    #
    def _start_sub_agents(self):
        '''
        For the list of sub_agents, get a launch command and launch that
        agent instance on the respective node.  We pass it to the seconds
        bootstrap level, there is no need to pass the first one again.
        '''

        # FIXME: reroute to agent daemonizer

        if not self._cfg.get('agents'):
            return

        self._log.debug('start_sub_agents')

        # the configs are written, and the sub-agents can be started.  To know
        # how to do that we create the agent launch method, have it creating
        # the respective command lines per agent instance, and run via
        # popen.
        #
        # actually, we only create the agent_lm once we really need it for
        # non-local sub_agents.
        agent_lm   = None
        for sa in self._cfg['agents']:

            target = self._cfg['agents'][sa]['target']

            if target == 'local':

                # start agent locally
                cmdline = '/bin/sh -l %s/bootstrap_2.sh %s' % (self._pwd, sa)

            elif target == 'node':

                if not agent_lm:
                    agent_lm = LaunchMethod.create(
                        name    = self._cfg['agent_launch_method'],
                        cfg     = self._cfg,
                        session = self._session)

                node = self._cfg['rm_info']['agent_nodes'][sa]
                # start agent remotely, use launch method
                # NOTE:  there is some implicit assumption that we can use
                #        the 'agent_node' string as 'agent_string:0' and
                #        obtain a well format slot...
                # FIXME: it is actually tricky to translate the agent_node
                #        into a viable 'slots' structure, as that is
                #        usually done by the schedulers.  So we leave that
                #        out for the moment, which will make this unable to
                #        work with a number of launch methods.  Can the
                #        offset computation be moved to the ResourceManager?
                bs_name = "%s/bootstrap_2.sh" % (self._pwd)
                ls_name = "%s/%s.sh" % (self._pwd, sa)
                slots = {
                    'cpu_processes'    : 1,
                    'cpu_threads'      : 1,
                    'gpu_processes'    : 0,
                    'gpu_threads'      : 0,
                  # 'nodes'            : [[node[0], node[1], [[0]], []]],
                    'nodes'            : [{'name'    : node[0],
                                           'uid'     : node[1],
                                           'core_map': [[0]],
                                           'gpu_map' : [],
                                           'lfs'     : {'path': '/tmp', 'size': 0}
                                         }],
                    'cores_per_node'   : self._cfg['rm_info']['cores_per_node'],
                    'gpus_per_node'    : self._cfg['rm_info']['gpus_per_node'],
                    'lm_info'          : self._cfg['rm_info']['lm_info'],
                }
                agent_cmd = {
                    'uid'              : sa,
                    'slots'            : slots,
                    'unit_sandbox_path': self._pwd,
                    'description'      : {'cpu_processes'    : 1,
                                          'gpu_process_type' : 'posix',
                                          'gpu_thread_type'  : 'posix',
                                          'executable'       : "/bin/sh",
                                          'mpi'              : False,
                                          'arguments'        : [bs_name, sa],
                                         }
                }
                cmd, hop = agent_lm.construct_command(agent_cmd,
                        launch_script_hop='/usr/bin/env RP_SPAWNER_HOP=TRUE "%s"' % ls_name)

                with open (ls_name, 'w') as ls:
                    # note that 'exec' only makes sense if we don't add any
                    # commands (such as post-processing) after it.
                    ls.write('#!/bin/sh\n\n')
                    for k,v in agent_cmd['description'].get('environment', {}).items():
                        ls.write('export "%s"="%s"\n' % (k, v))
                    ls.write('\n')
                    for pe_cmd in agent_cmd['description'].get('pre_exec', []):
                        ls.write('%s\n' % pe_cmd)
                    ls.write('\n')
                    ls.write('exec %s\n\n' % cmd)
                    st = os.stat(ls_name)
                    os.chmod(ls_name, st.st_mode | stat.S_IEXEC)

                if hop : cmdline = hop
                else   : cmdline = ls_name

            # ------------------------------------------------------------------
            class _SA(mp.Process):

                def __init__(self, sa, cmd, log):
                    self._name = sa
                    self._cmd  = cmd.split()
                    self._log  = log
                    self._proc = None
                    super(_SA, self).__init__(name=self._name)

                    self.start()


                def run(self):

                    sys.stdout = open('%s.out' % self._name, 'w')
                    sys.stderr = open('%s.err' % self._name, 'w')
                    out        = open('%s.out' % self._name, 'w')
                    err        = open('%s.err' % self._name, 'w')
                    self._proc = sp.Popen(args=self._cmd, stdout=out, stderr=err)
                    self._log.debug('sub-agent %s spawned [%s]', self._name,
                            self._proc)

                    assert(self._proc)

                    # FIXME: lifetime, use daemon agent launcher
                    while True:
                        time.sleep(0.1)
                        if self._proc.poll() is None:
                            return True   # all is well
                        else:
                            return False  # proc is gone - terminate
            # ------------------------------------------------------------------

            # spawn the sub-agent
            self._log.info ('create sub-agent %s: %s' % (sa, cmdline))
            _SA(sa, cmdline, log=self._log)

            # FIXME: register heartbeats?

        self._log.debug('start_sub_agents done')


    # --------------------------------------------------------------------------
    #
    def _agent_command_cb(self):

        if not self._check_commands(): return False
        if not self._check_state   (): return False

        return True


    # --------------------------------------------------------------------------
    #
    def _check_commands(self):

        # Check if there's a command waiting
        # FIXME: this pull should be done by the update worker, and commands
        #        should then be communicated over the command pubsub
        # FIXME: commands go to pmgr, umgr, session docs
        # FIXME: check if pull/wipe are atomic
        retdoc = self._dbs._c.find_and_modify(
                    query ={'uid'  : self._pid},
                    update={'$set' : {'cmd': []}},  # Wipe content of array
                    fields=['cmd'])

        if not retdoc:
            return True

        for spec in retdoc.get('cmd', []):

            cmd = spec['cmd']
            arg = spec['arg']

            self._prof.prof('cmd', msg="%s : %s" % (cmd, arg), uid=self._pid)

            if cmd == 'heartbeat' and arg['pmgr'] == self._pmgr:
                self._hb.beat(uid=self._pmgr)

            elif cmd == 'cancel_pilot':
                self._log.info('cancel pilot cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'terminate',
                                                  'arg' : None})
                self._final_cause = 'cancel'
                self.stop()

                return False  # we are done

            elif cmd == 'cancel_units':
                self._log.info('cancel_units cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_units',
                                                  'arg' : arg})
            else:
                self._log.warn('could not interpret cmd "%s" - ignore', cmd)

        return True


    # --------------------------------------------------------------------------
    #
    def _check_state(self):

        # Make sure that we haven't exceeded the runtime - otherwise terminate.
        if self._cfg.runtime:
            if time.time() >= self._starttime +  (int(self._cfg.runtime) * 60):
                self._log.info('runtime limit (%ss).', self._cfg.runtime * 60)
                self._final_cause = 'timeout'
                self.stop()
                return False  # we are done

        return True


    # --------------------------------------------------------------------------
    #
    def _check_units_cb(self):

        # Check for compute units waiting for input staging and log pull.
        #
        # FIXME: Unfortunately, 'find_and_modify' is not bulkable, so we have
        #        to use 'find'.  To avoid finding the same units over and over
        #        again, we update the 'control' field *before* running the next
        #        find -- so we do it right here.
        #        This also blocks us from using multiple ingest threads, or from
        #        doing late binding by unit pull :/
        unit_cursor = self._dbs._c.find({'type'    : 'unit',
                                         'pilot'   : self._pid,
                                         'control' : 'agent_pending'})
        if not unit_cursor.count():
            self._log.info('units pulled:    0')
            return True

        # update the units to avoid pulling them again next time.
        unit_list = list(unit_cursor)
        unit_uids = [unit['uid'] for unit in unit_list]

        self._dbs._c.update({'type'  : 'unit',
                             'uid'   : {'$in'     : unit_uids}},
                            {'$set'  : {'control' : 'agent'}},
                            multi=True)

        self._log.info("units pulled: %4d", len(unit_list))
        self._prof.prof('get', msg='bulk: %d' % len(unit_list), uid=self._pid)

        for unit in unit_list:

            # make sure the units obtain env settings (if needed)
            if 'unit_environment' in self._cfg:
                if not unit['description'].get('environment'):
                    unit['description']['environment'] = dict()
                for k,v in self._cfg['unit_environment'].items():
                    unit['description']['environment'][k] = v

            # we need to make sure to have the correct state:
            unit['state'] = rps._unit_state_collapse(unit['states'])
            self._prof.prof('get', uid=unit['uid'])

            # FIXME: raise or fail unit!
            if unit['state'] != rps.AGENT_STAGING_INPUT_PENDING:
                self._log.error('invalid state: %s', (pprint.pformat(unit)))

            unit['control'] = 'agent'

        # now we really own the CUs, and can start working on them (ie. push
        # them into the pipeline).  We don't publish nor profile as advance,
        # since that happened already on the module side when the state was set.
        self.advance(unit_list, publish=False, push=True)

        return True


# ------------------------------------------------------------------------------




import os
import sys


nfolders = int(sys.argv[1])

for i in range(nfolders):

    path = '/tmp/stage_in_folder_%d' % i

    if not os.path.exists(path):
        os.makedirs(path)

    with open('/tmp/stage_in_folder_%d/input_file.dat' % i, 'w') as fin:
        fin.write('hello world!')





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class IBRun(LaunchMethod):

    node_list = None

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        LaunchMethod.__init__(self, name, cfg, session)

        self._node_list = self._cfg.rm_info.node_list


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # ibrun: wrapper for mpirun at TACC
        self.launch_command = ru.which('ibrun')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']

        task_exec    = cud['executable']
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)
        task_env     = cud.get('environment') or dict()

        n_tasks      = cud['cpu_processes']

        # Usage of env variable TACC_TASKS_PER_NODE is purely for MPI tasks,
        #  and threads are not considered (info provided by TACC support)
        n_node_tasks = int(task_env.get('TACC_TASKS_PER_NODE') or
                           self._cfg.get('cores_per_node', 1))

        # TACC_TASKS_PER_NODE is used to set the actual number of running tasks,
        # if not set, then ibrun script will use the default slurm setting for
        # the number of tasks per node to build the hostlist (for TACC machines)
        #   NOTE: in case of performance issue please consider this parameter
        #   at the first place

        assert (slots.get('nodes') is not None), 'unit.slots.nodes is not set'

        ibrun_offset = 0
        offsets      = list()
        node_id      = 0

        for node in self._node_list:
            for slot_node in slots['nodes']:
                if slot_node['uid'] == node[0]:
                    for core_map in slot_node['core_map']:
                        assert core_map, 'core_map is not set'
                        # core_map contains core ids for each thread,
                        # but threads are ignored for offsets
                        offsets.append(node_id + (core_map[0] // len(core_map)))
            node_id += n_node_tasks

        if offsets:
            ibrun_offset = min(offsets)

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec

        ibrun_command = "%s -n %s -o %d %s" % \
                        (self.launch_command, n_tasks,
                         ibrun_offset, task_command)

        return ibrun_command, None


# ------------------------------------------------------------------------------




__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os

import radical.utils      as ru

from ... import states    as rps
from ... import constants as rpc
from ... import utils     as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's agent staging input types
RP_ASI_NAME_DEFAULT = "DEFAULT"


# ------------------------------------------------------------------------------
#
class AgentStagingInputComponent(rpu.Component):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id(cfg['owner'] + '.staging.input.%(counter)s',
                                   ru.ID_CUSTOM)

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Stager
    #
    @classmethod
    def create(cls, cfg, session):

        name = cfg.get('agent_staging_input_component', RP_ASI_NAME_DEFAULT)

        # Make sure that we are the base-class!
        if cls != AgentStagingInputComponent:
            raise TypeError("Factory only available to base class!")

        from .default import Default

        try:
            impl = {
                RP_ASI_NAME_DEFAULT: Default
            }[name]

            impl = impl(cfg, session)
            return impl

        except KeyError:
            raise ValueError("AgentStagingInputComponent '%s' unknown or defunct" % name)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_STAGING_INPUT_PENDING,
                            rpc.AGENT_STAGING_INPUT_QUEUE, self.work)

        self.register_output(rps.AGENT_SCHEDULING_PENDING,
                             rpc.AGENT_SCHEDULING_QUEUE)


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_STAGING_INPUT, publish=True, push=False)

        self._work(units)

    # --------------------------------------------------------------------------
    #
    def _work(self, units):

        # this needs to be overloaded by the inheriting implementation
        raise NotImplementedError('work() is not implemented')


# ------------------------------------------------------------------------------




# pylint: disable=subprocess-popen-preexec-fn
# FIXME: review pylint directive - https://github.com/PyCQA/pylint/pull/2087
#        (https://docs.python.org/3/library/subprocess.html#popen-constructor)

__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import stat
import time
import queue
import atexit
import pprint
import signal
import tempfile
import threading as mt
import traceback
import subprocess

import radical.utils as ru

from ...  import agent     as rpa
from ...  import utils     as rpu
from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
# ensure tasks are killed on termination
_pids = list()


def _kill():
    for pid in _pids:
        os.killpg(pid, signal.SIGTERM)


atexit.register(_kill)
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
#
class Popen(AgentExecutingComponent) :

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._watcher   = None
        self._terminate = mt.Event()

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        self._cancel_lock    = ru.RLock()
        self._cus_to_cancel  = list()
        self._cus_to_watch   = list()
        self._watch_queue    = queue.Queue ()

        self._pid = self._cfg['pid']

        # run watcher thread
        self._watcher = mt.Thread(target=self._watch)
      # self._watcher.daemon = True
        self._watcher.start()

        # The AgentExecutingComponent needs the LaunchMethod to construct
        # commands.
        self._task_launcher = rpa.LaunchMethod.create(
                name    = self._cfg.get('task_launch_method'),
                cfg     = self._cfg,
                session = self._session)

        self._mpi_launcher = rpa.LaunchMethod.create(
                name    = self._cfg.get('mpi_launch_method'),
                cfg     = self._cfg,
                session = self._session)

        self.gtod   = "%s/gtod" % self._pwd
        self.tmpdir = tempfile.gettempdir()


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_units':

            self._log.info("cancel_units command (%s)" % arg)
            with self._cancel_lock:
                self._cus_to_cancel.extend(arg['uids'])

        return True


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_EXECUTING, publish=True, push=False)

        for unit in units:
            self._handle_unit(unit)


    # --------------------------------------------------------------------------
    #
    def _handle_unit(self, cu):

        try:
            # prep stdout/err so that we can append w/o checking for None
            cu['stdout'] = ''
            cu['stderr'] = ''

            cpt = cu['description']['cpu_process_type']
          # gpt = cu['description']['gpu_process_type']  # FIXME: use

            # FIXME: this switch is insufficient for mixed units (MPI/OpenMP)
            if cpt == 'MPI': launcher = self._mpi_launcher
            else           : launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (process type = %s)" % cpt)

            self._log.debug("Launching unit with %s (%s).",
                            launcher.name, launcher.launch_command)

            # Start a new subprocess to launch the unit
            self.spawn(launcher=launcher, cu=cu)

        except Exception as e:
            # append the startup error to the units stderr.  This is
            # not completely correct (as this text is not produced
            # by the unit), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running CU")
            if cu.get('stderr') is None:
                cu['stderr'] = ''
            cu['stderr'] += "\nPilot cannot start compute unit:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            self._prof.prof('unschedule_start', uid=cu['uid'])
            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

            self.advance(cu, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, cu):

        descr   = cu['description']
        sandbox = cu['unit_sandbox_path']

        # make sure the sandbox exists
        self._prof.prof('exec_mkdir', uid=cu['uid'])
        rpu.rec_makedir(sandbox)
        self._prof.prof('exec_mkdir_done', uid=cu['uid'])

        launch_script_name = '%s/%s.sh' % (sandbox, cu['uid'])
        slots_fname        = '%s/%s.sl' % (sandbox, cu['uid'])

        self._log.debug("Created launch_script: %s", launch_script_name)

        # prep stdout/err so that we can append w/o checking for None
        cu['stdout'] = ''
        cu['stderr'] = ''

        with open(slots_fname, "w") as launch_script:
            launch_script.write('\n%s\n\n' % pprint.pformat(cu['slots']))

        with open(launch_script_name, "w") as launch_script:
            launch_script.write('#!/bin/sh\n\n')

            # Create string for environment variable setting
            env_string = ''
          # env_string += '. %s/env.orig\n'                % self._pwd
            env_string += 'export RADICAL_BASE="%s"\n'     % self._pwd
            env_string += 'export RP_SESSION_ID="%s"\n'    % self._cfg['sid']
            env_string += 'export RP_PILOT_ID="%s"\n'      % self._cfg['pid']
            env_string += 'export RP_AGENT_ID="%s"\n'      % self._cfg['aid']
            env_string += 'export RP_SPAWNER_ID="%s"\n'    % self.uid
            env_string += 'export RP_UNIT_ID="%s"\n'       % cu['uid']
            env_string += 'export RP_UNIT_NAME="%s"\n'     % cu['description'].get('name')
            env_string += 'export RP_GTOD="%s"\n'          % self.gtod
            env_string += 'export RP_TMP="%s"\n'           % self._cu_tmp
            env_string += 'export RP_PILOT_SANDBOX="%s"\n' % self._pwd
            env_string += 'export RP_PILOT_STAGING="%s/staging_area"\n' \
                                                           % self._pwd
            if self._prof.enabled:
                env_string += 'export RP_PROF="%s/%s.prof"\n' % (sandbox, cu['uid'])

            else:
                env_string += 'unset  RP_PROF\n'

            if 'RP_APP_TUNNEL' in os.environ:
                env_string += 'export RP_APP_TUNNEL="%s"\n' % os.environ['RP_APP_TUNNEL']

            env_string += '''
prof(){
    if test -z "$RP_PROF"
    then
        return
    fi
    event=$1
    msg=$2
    now=$($RP_GTOD)
    echo "$now,$event,unit_script,MainThread,$RP_UNIT_ID,AGENT_EXECUTING,$msg" >> $RP_PROF
}
'''

            # FIXME: this should be set by an LaunchMethod filter or something (GPU)
            env_string += 'export OMP_NUM_THREADS="%s"\n' % descr['cpu_threads']

            # The actual command line, constructed per launch-method
            try:
                launch_command, hop_cmd = launcher.construct_command(cu, launch_script_name)

                if hop_cmd : cmdline = hop_cmd
                else       : cmdline = launch_script_name

            except Exception as e:
                msg = "Error in spawner (%s)" % e
                self._log.exception(msg)
                raise RuntimeError (msg) from e

            # also add any env vars requested in the unit description
            if descr['environment']:
                for key,val in descr['environment'].items():
                    env_string += 'export "%s=%s"\n' % (key, val)

            launch_script.write('\n# Environment variables\n%s\n' % env_string)
            launch_script.write('prof cu_start\n')
            launch_script.write('\n# Change to unit sandbox\ncd %s\n' % sandbox)

            # Before the Big Bang there was nothing
            if self._cfg.get('cu_pre_exec'):
                for val in self._cfg['cu_pre_exec']:
                    launch_script.write("%s\n"  % val)

            if descr['pre_exec']:
                fail = ' (echo "pre_exec failed"; false) || exit'
                pre  = ''
                for elem in descr['pre_exec']:
                    pre += "%s || %s\n" % (elem, fail)
                # Note: extra spaces below are for visual alignment
                launch_script.write("\n# Pre-exec commands\n")
                launch_script.write('prof cu_pre_start\n')
                launch_script.write(pre)
                launch_script.write('prof cu_pre_stop\n')

            launch_script.write("\n# The command to run\n")
            launch_script.write('prof cu_exec_start\n')
            launch_script.write('%s\n' % launch_command)
            launch_script.write('RETVAL=$?\n')
            launch_script.write('prof cu_exec_stop\n')

            # After the universe dies the infrared death, there will be nothing
            if descr['post_exec']:
                fail = ' (echo "post_exec failed"; false) || exit'
                post = ''
                for elem in descr['post_exec']:
                    post += "%s || %s\n" % (elem, fail)
                launch_script.write("\n# Post-exec commands\n")
                launch_script.write('prof cu_post_start\n')
                launch_script.write('%s\n' % post)
                launch_script.write('prof cu_post_stop "$ret=RETVAL"\n')

            launch_script.write("\n# Exit the script with the return code from the command\n")
            launch_script.write("prof cu_stop\n")
            launch_script.write("exit $RETVAL\n")

        # done writing to launch script, get it ready for execution.
        st = os.stat(launch_script_name)
        os.chmod(launch_script_name, st.st_mode | stat.S_IEXEC)

        # prepare stdout/stderr
        stdout_file = descr.get('stdout') or '%s.out' % cu['uid']
        stderr_file = descr.get('stderr') or '%s.err' % cu['uid']

        cu['stdout_file'] = os.path.join(sandbox, stdout_file)
        cu['stderr_file'] = os.path.join(sandbox, stderr_file)

        _stdout_file_h = open(cu['stdout_file'], 'a')
        _stderr_file_h = open(cu['stderr_file'], 'a')

        self._log.info("Launching unit %s via %s in %s", cu['uid'], cmdline, sandbox)

        self._prof.prof('exec_start', uid=cu['uid'])
        cu['proc'] = subprocess.Popen(args       = cmdline,
                                      executable = None,
                                      stdin      = None,
                                      stdout     = _stdout_file_h,
                                      stderr     = _stderr_file_h,
                                      preexec_fn = os.setsid,
                                      close_fds  = True,
                                      shell      = True,
                                      cwd        = sandbox)
        self._prof.prof('exec_ok', uid=cu['uid'])

        # store pid for last-effort termination
        _pids.append(cu['proc'].pid)

        self._watch_queue.put(cu)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        try:
            while not self._terminate.is_set():

                cus = list()
                try:
                    # we don't want to only wait for one CU -- then we would
                    # pull CU state too frequently.  OTOH, we also don't want to
                    # learn about CUs until all slots are filled, because then
                    # we may not be able to catch finishing CUs in time -- so
                    # there is a fine balance here.  Balance means 100 (FIXME).
                    MAX_QUEUE_BULKSIZE = 100
                    while len(cus) < MAX_QUEUE_BULKSIZE :
                        cus.append (self._watch_queue.get_nowait())

                except queue.Empty:
                    # nothing found -- no problem, see if any CUs finished
                    pass

                # add all cus we found to the watchlist
                for cu in cus :
                    self._cus_to_watch.append (cu)

                # check on the known cus.
                action = self._check_running()

                if not action and not cus :
                    # nothing happened at all!  Zzz for a bit.
                    # FIXME: make configurable
                    time.sleep(0.1)

        except Exception as e:
            self._log.exception("Error in ExecWorker watch loop (%s)" % e)
            # FIXME: this should signal the ExecWorker for shutdown...


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self):

        action = 0
        for cu in self._cus_to_watch:

            # poll subprocess object
            exit_code = cu['proc'].poll()
            uid       = cu['uid']

            if exit_code is None:
                # Process is still running

                if cu['uid'] in self._cus_to_cancel:

                    # FIXME: there is a race condition between the state poll
                    # above and the kill command below.  We probably should pull
                    # state after kill again?

                    self._prof.prof('exec_cancel_start', uid=uid)

                    # We got a request to cancel this cu - send SIGTERM to the
                    # process group (which should include the actual launch
                    # method)
                  # cu['proc'].kill()
                    action += 1
                    try:
                        os.killpg(cu['proc'].pid, signal.SIGTERM)
                    except OSError:
                        # unit is already gone, we ignore this
                        pass
                    cu['proc'].wait()  # make sure proc is collected

                    with self._cancel_lock:
                        self._cus_to_cancel.remove(uid)

                    self._prof.prof('exec_cancel_stop', uid=uid)

                    del(cu['proc'])  # proc is not json serializable
                    self._prof.prof('unschedule_start', uid=cu['uid'])
                    self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)
                    self.advance(cu, rps.CANCELED, publish=True, push=False)

                    # we don't need to watch canceled CUs
                    self._cus_to_watch.remove(cu)

            else:

                self._prof.prof('exec_stop', uid=uid)

                # make sure proc is collected
                cu['proc'].wait()

                # we have a valid return code -- unit is final
                action += 1
                self._log.info("Unit %s has return code %s.", uid, exit_code)

                cu['exit_code'] = exit_code

                # Free the Slots, Flee the Flots, Ree the Frots!
                self._cus_to_watch.remove(cu)
                del(cu['proc'])  # proc is not json serializable
                self._prof.prof('unschedule_start', uid=cu['uid'])
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

                if exit_code != 0:
                    # The unit failed - fail after staging output
                    cu['target_state'] = rps.FAILED

                else:
                    # The unit finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there are no
                    # directives -- at the very least, we'll upload stdout/stderr
                    cu['target_state'] = rps.DONE

                self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, publish=True, push=True)

        return action


# ------------------------------------------------------------------------------




#!/usr/bin/env python

# This is an example MPI4Py program that is used
# by different examples and tests.

import time

from mpi4py import MPI


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    host  = MPI.Get_processor_name().split('.')[0]
    comm  = MPI.COMM_WORLD

    time.sleep(1)

    print("%d/%d/%s"  % (comm.rank + 1, comm.size, host))

    comm.Barrier()   # wait for everybody to synchronize here

# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__ = "MIT"

import math as m

from ...   import constants as rpc
from .base import AgentSchedulingComponent


# ------------------------------------------------------------------------------
#
# This is a continuous scheduler with awareness of a node's file-storage and
# memory capabilities.  The scheduler respects node tagging (place unit on same
# node as other units with the same tag).
#
#
# Continuous:
#
# The scheduler attempts to find continuous stretches of nodes to place
# a multinode-unit.
#
#
# Local Storage:
#
# The storage availability will be obtained from the rm_node_list and assigned
# to the node list of the class. The lfs requirement will be obtained from the
# cud in the alloc_nompi and alloc_mpi methods. Using the availability and
# requirement, the _find_resources method will return the core and gpu ids.
#
# Expected DS of the nodelist
# self.nodes = [{
#                   'name'    : 'aa',
#                   'uid'     : 'node.0000',
#                   'cores'   : [0, 1, 2, 3, 4, 5, 6, 7],
#                   'gpus'    : [0,1, 2],
#                   'lfs'     : 128,
#                   'mem'     : 256
#               },
#               {
#                   'name'    : 'bb',
#                   'uid'     : 'node.0001',
#                   'cores'   : [0, 1, 2, 3, 4, 5, 6, 7],
#                   'gpus'    : [0,1, 2],
#                   'lfs'     : 256,
#                   'mem'     : 256,
#                },
#                ...
#              ]
#
# lfs storage and memory is specified in MByte.  The scheduler assumes that
# both are freed when the unit finishes.
#
#
# Unit Tagging:
#
# The scheduler attempts to schedule units with the same tag onto the same node,
# so that the unit can reuse the previous unit's data.  This assumes that
# storage is not freed when the units finishes.
#
# FIXME: the alert reader will realize a discrepancy in the above set of
#        assumptions.
#
# ------------------------------------------------------------------------------
#
class Continuous(AgentSchedulingComponent):
    '''
    The Continuous scheduler attempts to place threads and processes of
    a compute units onto nodes in the cluster.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentSchedulingComponent.__init__(self, cfg, session)

        self._tag_history   = dict()
        self._scattered     = None
        self._node_offset   = 0


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # * scattered:
        #   This is the continuous scheduler, because it attempts to allocate
        #   a *continuous* set of cores/nodes for a unit.  It does, however,
        #   also allow to scatter the allocation over discontinuous nodes if
        #   this option is set.  This implementation is not optimized for the
        #   scattered mode!  The default is 'False'.
        #
        self._scattered = self._cfg.get('scattered', False)

        self.nodes = []
        for node, node_uid in self._rm_node_list:

            node_entry = {'name'   : node,
                          'uid'    : node_uid,
                          'cores'  : [rpc.FREE] * self._rm_cores_per_node,
                          'gpus'   : [rpc.FREE] * self._rm_gpus_per_node,
                          'lfs'    :              self._rm_lfs_per_node,
                          'mem'    :              self._rm_mem_per_node}

            # summit
            if  self._rm_cores_per_node > 40 and \
                self._cfg['task_launch_method'] == 'JSRUN':

                # Summit cannot address the last core of the second socket at
                # the moment, so we mark it as `DOWN` and the scheduler skips
                # it.  We need to check the SMT setting to make sure the right
                # logical cores are marked.  The error we see on those cores is:
                # "ERF error: 1+ cpus are not available"
                #
                # This is related to the known issue listed on
                # https://www.olcf.ornl.gov/for-users/system-user-guides \
                #                          /summit/summit-user-guide/
                #
                # "jsrun explicit resource file (ERF) allocates incorrect
                # resources"
                #
                smt = self._rm_info.get('smt', 1)

                # only socket `1` is affected at the moment
              # for s in [0, 1]:
                for s in [1]:
                    for i in range(smt):
                        idx = s * 21 * smt + i
                        node_entry['cores'][idx] = rpc.DOWN

            self.nodes.append(node_entry)

        if self._rm_cores_per_node > 40 and \
           self._cfg['task_launch_method'] == 'JSRUN':
            self._rm_cores_per_node -= 1


    # --------------------------------------------------------------------------
    #
    def _iterate_nodes(self):
        # note that the first index is yielded twice, so that the respecitve
        # node can function as first and last node in an allocation.

        iterator_count = 0

        while iterator_count < len(self.nodes):
            yield self.nodes[self._node_offset]
            iterator_count    += 1
            self._node_offset += 1
            self._node_offset  = self._node_offset % len(self.nodes)


    # --------------------------------------------------------------------------
    #
    def unschedule_unit(self, unit):
        '''
        This method is called when previously aquired resources are not needed
        anymore.  `slots` are the resource slots as previously returned by
        `schedule_unit()`.
        '''

        # reflect the request in the nodelist state (set to `FREE`)
        self._change_slot_states(unit['slots'], rpc.FREE)


    # --------------------------------------------------------------------------
    #
    def _find_resources(self, node, find_slots, cores_per_slot, gpus_per_slot,
                              lfs_per_slot, mem_per_slot, partial):
        '''
        Find up to the requested number of slots, where each slot features the
        respective `x_per_slot` resources.  This call will return a list of
        slots of the following structure:

            {
                'node_name': 'node_1',
                'node_uid' : 'node_1',
                'cores'    : [1, 2, 4, 5],
                'gpus'     : [1, 3],
                'lfs'      : 1234,
                'mem'      : 4321
            }

        The call will *not* change the allocation status of the node, atomicity
        must be guaranteed by the caller.

        We don't care about continuity within a single node - cores `[1, 5]` are
        assumed to be as close together as cores `[1, 2]`.

        When `partial` is set, the method CAN return less than `find_slots`
        number of slots - otherwise, the method returns the requested number of
        slots or `None`.

        FIXME: SMT handling: we should assume that hardware threads of the same
               physical core cannot host different executables, so HW threads
               can only account for thread placement, not process placement.
               This might best be realized by internally handling SMT as minimal
               thread count and using physical core IDs for process placement?
        '''

        # check if the node can host the request
        free_cores = node['cores'].count(rpc.FREE)
        free_gpus  = node['gpus'].count(rpc.FREE)
        free_lfs   = node['lfs']['size']
        free_mem   = node['mem']

        # check how many slots we can serve, at most
        alc_slots = 1
        if cores_per_slot:
            alc_slots = int(m.floor(free_cores / cores_per_slot))

        if gpus_per_slot:
            alc_slots = min(alc_slots, int(m.floor(free_gpus / gpus_per_slot )))

        if lfs_per_slot:
            alc_slots = min(alc_slots, int(m.floor(free_lfs / lfs_per_slot )))

        if mem_per_slot:
            alc_slots = min(alc_slots, int(m.floor(free_mem / mem_per_slot )))

        # is this enough?
        if not alc_slots:
            return None

        if not partial:
            if alc_slots < find_slots:
                return None

        # find at most `find_slots`
        alc_slots = min(alc_slots, find_slots)

        # we should be able to host the slots - dig out the precise resources
        slots     = list()
        node_uid  = node['uid']
        node_name = node['name']

        core_idx  = 0
        gpu_idx   = 0

        for _ in range(alc_slots):

            cores = list()
            gpus  = list()

            while len(cores) < cores_per_slot:

                if node['cores'][core_idx] == rpc.FREE:
                    cores.append(core_idx)
                core_idx += 1

            while len(gpus) < gpus_per_slot:

                if node['gpus'][gpu_idx] == rpc.FREE:
                    gpus.append(gpu_idx)
                gpu_idx += 1

            core_map = [cores]
            gpu_map  = [[gpu] for gpu in gpus]

            slots.append({'uid'     : node_uid,
                          'name'    : node_name,
                          'core_map': core_map,
                          'gpu_map' : gpu_map,
                          'lfs'     : {'size': lfs_per_slot,
                                       'path': self._rm_lfs_per_node['path']},
                          'mem'     : mem_per_slot})

        # consistency check
        assert((len(slots) == find_slots) or (len(slots) and partial))

        return slots


    # --------------------------------------------------------------------------
    #
    #
    def schedule_unit(self, unit):
        '''
        Find an available set of slots, potentially across node boundaries (in
        the MPI case).  By default, we only allow for partial allocations on the
        first and last node - but all intermediate nodes MUST be completely used
        (this is the 'CONTINUOUS' scheduler after all).

        If the scheduler is configured with `scattered=True`, then that
        constraint is relaxed, and any set of slots (be it continuous across
        nodes or not) is accepted as valid allocation.

        No matter the mode, we always make sure that we allocate slots in chunks
        of cores, gpus, lfs and mem required per process - otherwise the
        application processes would not be able to acquire the requested
        resources on the respective node.

        Contrary to the documentation, this scheduler interprets `gpu_processes`
        as number of GPUs that need to be available to each process, i.e., as
        `gpus_per_process'.

        Note that all resources for non-MPI tasks will always need to be placed
        on a single node.
        '''

        self._log.debug('find_resources %s', unit['uid'])

        cud = unit['description']
        mpi = bool('mpi' in cud['cpu_process_type'].lower())

        # dig out the allocation request details
        req_slots      = cud['cpu_processes']
        cores_per_slot = cud['cpu_threads']
        gpus_per_slot  = cud['gpu_processes']
        lfs_per_slot   = cud['lfs_per_process']
        mem_per_slot   = cud['mem_per_process']

        # make sure that processes are at least single-threaded
        if not cores_per_slot:
            cores_per_slot = 1

        self._log.debug('req : %s %s %s %s %s', req_slots, cores_per_slot,
                        gpus_per_slot, lfs_per_slot, mem_per_slot)

        # First and last nodes can be a partial allocation - all other nodes
        # can only be partial when `scattered` is set.
        #
        # Iterate over all nodes until we find something. Check if it fits the
        # allocation mode and sequence.  If not, start over with the next node.
        # If it matches, add the slots found and continue to next node.
        #
        # FIXME: persistent node index

        cores_per_node = self._rm_cores_per_node
        gpus_per_node  = self._rm_gpus_per_node
        lfs_per_node   = self._rm_lfs_per_node['size']
        mem_per_node   = self._rm_mem_per_node

        # we always fail when too many threads are requested
        assert(cores_per_slot <= cores_per_node), 'too many threads per proc %s' % cores_per_slot
        assert(gpus_per_slot  <= gpus_per_node),  'too many gpus    per proc %s' % gpus_per_slot
        assert(lfs_per_slot   <= lfs_per_node),   'too much lfs     per proc %s' % lfs_per_slot
        assert(mem_per_slot   <= mem_per_node),   'too much mem     per proc %s' % mem_per_slot

        # check what resource type limits teh number of slots per node
        slots_per_node = int(m.floor(cores_per_node / cores_per_slot))

        if gpus_per_slot:
            slots_per_node = min(slots_per_node,
                                 int(m.floor(gpus_per_node / gpus_per_slot )))

        if lfs_per_slot:
            slots_per_node = min(slots_per_node,
                                 int(m.floor(lfs_per_node / lfs_per_slot)))

        if mem_per_slot:
            slots_per_node = min(slots_per_node,
                                 int(m.floor(mem_per_node / mem_per_slot)))

        if not mpi and req_slots > slots_per_node:
            raise ValueError('non-mpi task does not fit on a single node')

        # set conditions to find the first matching node
        is_first = True
        is_last  = False
        tag      = cud.get('tag')

        # what remains to be allocated?  all of it right now.
        alc_slots = list()
        rem_slots = req_slots

        # start the search
        for node in self._iterate_nodes():

            node_uid  = node['uid']
          # node_name = node['name']

          # self._log.debug('next %s : %s', node_uid, node_name)
          # self._log.debug('req1: %s = %s + %s', req_slots, rem_slots,
          #                                       len(alc_slots))

            # Check if a unit is tagged to use this node.  This means we check
            #   - if a tag exists
            #   - if the tag has been used before
            #   - if the previous use included this node
            # If a tag exists, continue to consider this node if the tag was
            # used for this node - else continune to the next node.
            if tag:
                if tag in self._tag_history:
                    if node_uid not in self._tag_history[tag]:
                        continue

            # if only a small set of cores/gpus remains unallocated (ie. less
            # than node size), we are in fact looking for the last node.  Note
            # that this can also be the first node, for small units.
            if  rem_slots < slots_per_node:
                is_last = True

            # we allow partial nodes on the first and last node, and on any
            # node if a 'scattered' allocation is requested.
            if is_first or self._scattered or is_last:
                partial = True
            else:
                partial = False

            # but also, non-mpi tasks are never partially allocated
            if not mpi:
                partial = False

            # now we know how many slots we still need at this point - but
            # we only search up to node-size on this node.  Duh!
            find_slots = min(rem_slots, slots_per_node)
          # self._log.debug('find: %s', find_slots)

            # under the constraints so derived, check what we find on this node
            new_slots = self._find_resources(node           = node,
                                             find_slots     = find_slots,
                                             cores_per_slot = cores_per_slot,
                                             gpus_per_slot  = gpus_per_slot,
                                             lfs_per_slot   = lfs_per_slot,
                                             mem_per_slot   = mem_per_slot,
                                             partial        = partial)

            if not new_slots:

                # this was not a match. If we are in  'scattered' mode, we just
                # ignore this node.  Otherwise we have to restart the search
                # (continuity is broken)
                if not self._scattered:
                    alc_slots       = list()
                    rem_slots       = req_slots
                    is_first        = True
                    is_last         = False

                # try next node
                continue

            # this node got a match, store away the found slots and continue
            # search for remaining ones
            rem_slots -= len(new_slots)
            alc_slots.extend(new_slots)

          # self._log.debug('new slots: %s', pprint.pformat(new_slots))
          # self._log.debug('req2: %s = %s + %s <> %s', req_slots, rem_slots,
          #                                       len(new_slots), len(alc_slots))

            # we are young only once.  kinda...
            is_first = False

            # or maybe don't continue the search if we have in fact enough!
            if rem_slots == 0:
                break

        # if we did not find enough, there is not much we can do at this point
        if  rem_slots > 0:
            return None  # signal failure

        slots = {'nodes'         : alc_slots,
                 'cores_per_node': self._rm_cores_per_node,
                 'gpus_per_node' : self._rm_gpus_per_node,
                 'lfs_per_node'  : self._rm_lfs_per_node,
                 'mem_per_node'  : self._rm_mem_per_node,
                 'lm_info'       : self._rm_lm_info,
                }


        # allocation worked!  If the unit was tagged, store the node IDs for
        # this tag, so that later units can reuse that information
        if tag:
            self._tag_history[tag] = [node['uid'] for node in slots['nodes']]

        # this should be nicely filled out now - return
        return slots


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import fractions
import collections

import radical.utils as ru
from functools import reduce


# 'enum' for launch method types
LM_NAME_APRUN         = 'APRUN'
LM_NAME_CCMRUN        = 'CCMRUN'
LM_NAME_FORK          = 'FORK'
LM_NAME_IBRUN         = 'IBRUN'
LM_NAME_MPIEXEC       = 'MPIEXEC'
LM_NAME_MPIRUN        = 'MPIRUN'
LM_NAME_MPIRUN_MPT    = 'MPIRUN_MPT'
LM_NAME_MPIRUN_CCMRUN = 'MPIRUN_CCMRUN'
LM_NAME_MPIRUN_DPLACE = 'MPIRUN_DPLACE'
LM_NAME_MPIRUN_RSH    = 'MPIRUN_RSH'
LM_NAME_JSRUN         = 'JSRUN'
LM_NAME_PRTE          = 'PRTE'
LM_NAME_PRTE2         = 'PRTE2'
LM_NAME_FLUX          = 'FLUX'
LM_NAME_ORTE          = 'ORTE'
LM_NAME_ORTE_LIB      = 'ORTE_LIB'
LM_NAME_RSH           = 'RSH'
LM_NAME_SSH           = 'SSH'
LM_NAME_YARN          = 'YARN'
LM_NAME_SPARK         = 'SPARK'
LM_NAME_SRUN          = 'SRUN'

# deprecated
# LM_NAME_POE           = 'POE'
# LM_NAME_DPLACE        = 'DPLACE'
# LM_NAME_RUNJOB        = 'RUNJOB'

# deprecated
# LM_NAME_POE           = 'POE'
# LM_NAME_DPLACE        = 'DPLACE'
# LM_NAME_RUNJOB        = 'RUNJOB'

PWD = os.getcwd()


# ------------------------------------------------------------------------------
#
class LaunchMethod(object):

    # List of environment variables that designated Launch Methods should export
    # FIXME: we should find out what env vars are changed or added by
    #        cud.pre_exec, and then should also export those.  That would make
    #        our launch script ore complicated though...
    EXPORT_ENV_VARIABLES = [
        'LD_LIBRARY_PATH',
        'PATH',
        'PYTHONPATH',
        'OMP_NUM_THREADS',
        'CUDA_VISIBLE_DEVICES',
        'RP_AGENT_ID',
        'RP_GTOD',
        'RP_PILOT_ID',
        'RP_PILOT_STAGING',
        'RP_PROF',
        'RP_SESSION_ID',
        'RP_SPAWNER_ID',
        'RP_TMP',
        'RP_UNIT_ID',
        'RP_UNIT_NAME',
        'RP_PILOT_SANDBOX',
        'RADICAL_BASE'
    ]

    MPI_FLAVOR_OMPI    = 'OMPI'
    MPI_FLAVOR_HYDRA   = 'HYDRA'
    MPI_FLAVOR_UNKNOWN = 'unknown'

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, cfg, session):

        self.name     = name
        self._cfg     = cfg
        self._session = session
        self._log     = self._session._log    # pylint: disable=protected-access
        self._log.debug('create LaunchMethod: %s', type(self))

        # A per-launch_method list of env vars to remove from the CU env
        self.env_removables = []

        self._configure()


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Launch Method.
    #
    @classmethod
    def create(cls, name, cfg, session):

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod create only available to base class!")

        # In case of undefined LM just return None
        if not name:
            return None

        from .aprun          import APRun
        from .ccmrun         import CCMRun
        from .fork           import Fork
        from .ibrun          import IBRun
        from .mpiexec        import MPIExec
        from .mpirun         import MPIRun
        from .jsrun          import JSRUN
        from .prte           import PRTE
        from .prte2          import PRTE2
        from .flux           import Flux
        from .rsh            import RSH
        from .ssh            import SSH
        from .yarn           import Yarn
        from .spark          import Spark
        from .srun           import Srun

      # # deprecated
      # from .orte           import ORTE
      # from .orte_lib       import ORTELib
      # from .mpirun_ccmrun  import MPIRunCCMRun
      # from .mpirun_dplace  import MPIRunDPlace
      # from .mpirun_mpt     import MPIRun_MPT
      # from .mpirun_rsh     import MPIRunRSH
      # from .dplace         import DPlace
      # from .poe            import POE
      # from .runjob         import Runjob
      # from .mpirun_ccmrun  import MPIRunCCMRun
      # from .mpirun_dplace  import MPIRunDPlace
      # from .mpirun_mpt     import MPIRun_MPT
      # from .mpirun_rsh     import MPIRunRSH
      # from .dplace         import DPlace
      # from .poe            import POE
      # from .runjob         import Runjob

        try:
            impl = {
                LM_NAME_APRUN         : APRun,
                LM_NAME_CCMRUN        : CCMRun,
                LM_NAME_FORK          : Fork,
                LM_NAME_IBRUN         : IBRun,
                LM_NAME_MPIEXEC       : MPIExec,
                LM_NAME_MPIRUN        : MPIRun,
                LM_NAME_MPIRUN_CCMRUN : MPIRun,
                LM_NAME_MPIRUN_RSH    : MPIRun,
                LM_NAME_MPIRUN_MPT    : MPIRun,
                LM_NAME_MPIRUN_DPLACE : MPIRun,
                LM_NAME_JSRUN         : JSRUN,
                LM_NAME_PRTE          : PRTE,
                LM_NAME_PRTE2         : PRTE2,
                LM_NAME_FLUX          : Flux,
                LM_NAME_RSH           : RSH,
                LM_NAME_SSH           : SSH,
                LM_NAME_YARN          : Yarn,
                LM_NAME_SPARK         : Spark,
                LM_NAME_SRUN          : Srun,

              # # deprecated
              # LM_NAME_ORTE          : ORTE,
              # LM_NAME_ORTE_LIB      : ORTELib,
              # LM_NAME_DPLACE        : DPlace,
              # LM_NAME_POE           : POE,
              # LM_NAME_RUNJOB        : Runjob,
            }[name]
            return impl(name, cfg, session)

        except KeyError:
            session._log.exception('invalid lm %s' % name)
            return None

        except Exception:
            session._log.exception('unusable lm %s' % name)
            return None


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_config_hook(cls, name, cfg, rm, log, profiler):
        '''
        This hook will allow the ResourceManager to perform launch methods
        specific configuration steps.  The ResourceManager layer MUST ensure
        that this hook is called exactly once (globally).  This will be a NOOP
        for LaunchMethods which do not overload this method.  Exceptions fall
        through to the ResourceManager.
        '''

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod config hook only available to base class!")

        from .fork           import Fork
        from .prte           import PRTE
        from .prte2          import PRTE2
        from .flux           import Flux
        from .jsrun          import JSRUN
        from .yarn           import Yarn
        from .spark          import Spark

        # # deprecated
        # from .orte           import ORTE

        impl = {
            LM_NAME_FORK          : Fork,
            LM_NAME_PRTE          : PRTE,
            LM_NAME_PRTE2         : PRTE2,
            LM_NAME_FLUX          : Flux,
            LM_NAME_JSRUN         : JSRUN,
            LM_NAME_YARN          : Yarn,
            LM_NAME_SPARK         : Spark,

            # # deprecated
            # LM_NAME_ORTE          : ORTE,

        }.get(name)

        if not impl:
            log.info('no config hook defined for LaunchMethod %s' % name)
            return None

        log.info('ResourceManager config hook for LaunchMethod %s: %s' % (name, impl))
        return impl.rm_config_hook(name, cfg, rm, log, profiler)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def rm_shutdown_hook(cls, name, cfg, rm, lm_info, log, profiler):
        '''
        This hook is symmetric to the config hook above, and is called during
        shutdown sequence, for the sake of freeing allocated resources.
        '''

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod shutdown hook only available to base class!")

        from .prte           import PRTE
        from .prte2          import PRTE2
        from .flux           import Flux
        from .yarn           import Yarn
        from .spark          import Spark

        # # deprecated
        # from .orte           import ORTE

        impl = {
            LM_NAME_PRTE          : PRTE,
            LM_NAME_PRTE2         : PRTE2,
            LM_NAME_FLUX          : Flux,
            LM_NAME_YARN          : Yarn,
            LM_NAME_SPARK         : Spark

            # # deprecated
            # LM_NAME_ORTE          : ORTE,
        }.get(name)

        if not impl:
            log.info('no shutdown hook defined for LaunchMethod %s' % name)
            return None

        log.info('ResourceManager shutdown hook for LaunchMethod %s: %s' % (name, impl))
        return impl.rm_shutdown_hook(name, cfg, rm, lm_info, log, profiler)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _create_hostfile(cls, sandbox, uid, all_hosts, separator=' ',
                         impaired=False):

        # Open appropriately named temporary file
        # NOTE: we make an assumption about the unit sandbox here
        filename = '%s/%s.hosts' % (sandbox, uid)
        with open(filename, 'w') as fout:

            if not impaired:
                # Write "hostN x\nhostM y\n" entries
                # Create a {'host1': x, 'host2': y} dict
                counter = collections.Counter(all_hosts)

                # Convert it into an ordered dict,
                # which hopefully resembles the original ordering
                count_dict = collections.OrderedDict(sorted(counter.items(),
                                                     key=lambda t: t[0]))

                for (host, count) in count_dict.items():
                    fout.write('%s%s%d\n' % (host, separator, count))

            else:
                # Write "hostN\nhostM\n" entries
                for host in all_hosts:
                    fout.write('%s\n' % host)

        # Return the filename, caller is responsible for cleaning up
        return filename


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _compress_hostlist(cls, all_hosts):

        # Return gcd of a list of numbers
        def gcd_list(l):
            return reduce(fractions.gcd, l)

        # Create a {'host1': x, 'host2': y} dict
        count_dict = dict(collections.Counter(all_hosts))
        # Find the gcd of the host counts
        host_gcd = gcd_list(set(count_dict.values()))

        # Divide the host counts by the gcd
        for host in count_dict:
            count_dict[host] /= host_gcd

        # Recreate a list of hosts based on the normalized dict
        hosts = list()
        for (host, count) in list(count_dict.items()):
            hosts.extend([host] * count)

        # sort the list for readbility
        hosts.sort()

        return hosts


    # --------------------------------------------------------------------------
    #
    def _create_arg_string(self, args):

        # unit Arguments (if any)
        arg_string = ''
        if args:
            for arg in args:
                if not arg:
                    # ignore empty args
                    continue

                if arg in ['>', '>>', '<', '<<', '|', '||', '&&', '&']:
                    # Don't quote shell direction arguments, etc.
                    arg_string += '%s ' % arg
                    continue

                if any([c in arg for c in ['?', '*']]):
                    # Don't quote arguments with wildcards
                    arg_string += '%s ' % arg
                    continue

                arg = arg.replace('"', '\\"')    # Escape all double quotes
                if arg[0] == arg[-1] == "'" :    # between outer single quotes?
                    arg_string += '%s ' % arg    # ... pass it as is.

                else:
                    arg_string += '"%s" ' % arg  # else return double quoted

        return arg_string


    # --------------------------------------------------------------------------
    #
    def _get_mpi_info(self, exe):
        '''
        returns version and flavor of MPI version.
        '''

        if not exe:
            raise ValueError('no executable found')

        version = None
        flavor  = self.MPI_FLAVOR_UNKNOWN

        out, _, ret = ru.sh_callout('%s -v' % exe)

        if ret:
            out, _, ret = ru.sh_callout('%s --version' % exe)

        if ret:
            out, _, ret = ru.sh_callout('%s -info' % exe)

        if not ret:
            for line in out.splitlines():
                if 'hydra build details:' in line.lower():
                    version = line.split(':', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_HYDRA
                    break

                if 'mvapich2' in line.lower():
                    version = line
                    flavor  = self.MPI_FLAVOR_HYDRA
                    break

                if 'version:' in line.lower():
                    version = line.split(':', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_OMPI
                    break

                if '(open mpi)' in line.lower():
                    version = line.split(')', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_OMPI
                    break

        if not flavor:
            raise RuntimeError('cannot identify MPI flavor [%s]' % exe)

        self._log.debug('mpi version: %s [%s]', version, flavor)

        return version, flavor


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import time
import threading as mt

import radical.utils as ru

from . import utils     as rpu
from . import states    as rps
from . import constants as rpc

from . import compute_unit_description as rpcud


# bulk callbacks are implemented, but are currently not used nor exposed.
_USE_BULK_CB = False
if os.environ.get('RADICAL_PILOT_BULK_CB', '').lower() in ['true', 'yes', '1']:
    _USE_BULK_CB = True


# ------------------------------------------------------------------------------
#
class UnitManager(rpu.Component):
    """
    A UnitManager manages :class:`radical.pilot.ComputeUnit` instances which
    represent the **executable** workload in RADICAL-Pilot. A UnitManager
    connects the ComputeUnits with one or more :class:`Pilot` instances (which
    represent the workload **executors** in RADICAL-Pilot) and a **scheduler**
    which determines which :class:`ComputeUnit` gets executed on which
    :class:`Pilot`.

    **Example**::

        s = rp.Session(database_url=DBURL)

        pm = rp.PilotManager(session=s)

        pd = rp.ComputePilotDescription()
        pd.resource = "futuregrid.alamo"
        pd.cores = 16

        p1 = pm.submit_pilots(pd) # create first pilot with 16 cores
        p2 = pm.submit_pilots(pd) # create second pilot with 16 cores

        # Create a workload of 128 '/bin/sleep' compute units
        compute_units = []
        for unit_count in range(0, 128):
            cu = rp.ComputeUnitDescription()
            cu.executable = "/bin/sleep"
            cu.arguments = ['60']
            compute_units.append(cu)

        # Combine the two pilots, the workload and a scheduler via
        # a UnitManager.
        um = rp.UnitManager(session=session, scheduler=rp.SCHEDULER_ROUND_ROBIN)
        um.add_pilot(p1)
        um.submit_units(compute_units)


    The unit manager can issue notification on unit state changes.  Whenever
    state notification arrives, any callback registered for that notification is
    fired.

    NOTE: State notifications can arrive out of order wrt the unit state model!
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, session, cfg='default', scheduler=None):
        """
        Creates a new UnitManager and attaches it to the session.

        **Arguments:**
            * session [:class:`radical.pilot.Session`]:
              The session instance to use.
            * cfg (`dict` or `string`):
              The configuration or name of configuration to use.
            * scheduler (`string`):
              The name of the scheduler plug-in to use.

        **Returns:**
            * A new `UnitManager` object [:class:`radical.pilot.UnitManager`].
        """

        self._pilots      = dict()
        self._pilots_lock = ru.RLock('umgr.pilots_lock')
        self._units       = dict()
        self._units_lock  = ru.RLock('umgr.units_lock')
        self._callbacks   = dict()
        self._cb_lock     = ru.RLock('umgr.cb_lock')
        self._terminate   = mt.Event()
        self._closed      = False
        self._rec_id      = 0       # used for session recording
        self._uid         = ru.generate_id('umgr.%(item_counter)04d',
                                           ru.ID_CUSTOM, ns=session.uid)

        for m in rpc.UMGR_METRICS:
            self._callbacks[m] = dict()

        # NOTE: `name` and `cfg` are overloaded, the user cannot point to
        #       a predefined config and amed it at the same time.  This might
        #       be ok for the session, but introduces a minor API inconsistency.
        #
        name = None
        if isinstance(cfg, str):
            name = cfg
            cfg  = None

        cfg           = ru.Config('radical.pilot.umgr', name=name, cfg=cfg)
        cfg.uid       = self._uid
        cfg.owner     = self._uid
        cfg.sid       = session.uid
        cfg.base      = session.base
        cfg.path      = session.path
        cfg.dburl     = session.dburl
        cfg.heartbeat = session.cfg.heartbeat

        if scheduler:
            # overwrite the scheduler from the config file
            cfg.scheduler = scheduler


        rpu.Component.__init__(self, cfg, session=session)
        self.start()

        self._log.info('started umgr %s', self._uid)
        self._rep.info('<<create unit manager')

        # create pmgr bridges and components, use session cmgr for that
        self._cmgr = rpu.ComponentManager(self._cfg)
        self._cmgr.start_bridges()
        self._cmgr.start_components()

        # The output queue is used to forward submitted units to the
        # scheduling component.
        self.register_output(rps.UMGR_SCHEDULING_PENDING,
                             rpc.UMGR_SCHEDULING_QUEUE)

        # the umgr will also collect units from the agent again, for output
        # staging and finalization
        if self._cfg.bridges.umgr_staging_output_queue:
            self._has_sout = True
            self.register_output(rps.UMGR_STAGING_OUTPUT_PENDING,
                                 rpc.UMGR_STAGING_OUTPUT_QUEUE)
        else:
            self._has_sout = False

        # register the state notification pull cb
        # FIXME: this should be a tailing cursor in the update worker
        self.register_timed_cb(self._state_pull_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # register callback which pulls units back from agent
        # FIXME: this should be a tailing cursor in the update worker
        self.register_timed_cb(self._unit_pull_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # also listen to the state pubsub for unit state changes
        self.register_subscriber(rpc.STATE_PUBSUB, self._state_sub_cb)

        # let session know we exist
        self._session._register_umgr(self)

        self._prof.prof('setup_done', uid=self._uid)
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # the manager must not carry bridge and component handles across forks
        ru.atfork(self._atfork_prepare, self._atfork_parent, self._atfork_child)


    # --------------------------------------------------------------------------
    #
    # EnTK forks, make sure we don't carry traces of children across the fork
    #
    def _atfork_prepare(self): pass
    def _atfork_parent(self) : pass
    def _atfork_child(self)  :
        self._bridges    = dict()
        self._components = dict()


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self._cmgr.close()


    # --------------------------------------------------------------------------
    #
    def close(self):
        """
        Shut down the UnitManager and all its components.
        """

        # we do not cancel units at this point, in case any component or pilot
        # wants to continue to progress unit states, which should indeed be
        # independent from the umgr life cycle.

        if self._closed:
            return

        self._terminate.set()
        self._rep.info('<<close unit manager')

        # disable callbacks during shutdown
        with self._cb_lock:
            self._callbacks = dict()
            for m in rpc.UMGR_METRICS:
                self._callbacks[m] = dict()

        self._cmgr.close()

        self._log.info("Closed UnitManager %s." % self._uid)

        self._closed = True
        self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """
        Returns a dictionary representation of the UnitManager object.
        """

        ret = {
            'uid': self.uid,
            'cfg': self.cfg
        }

        return ret


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        """
        Returns a string representation of the UnitManager object.
        """

        return str(self.as_dict())


    # --------------------------------------------------------------------------
    #
    def _pilot_state_cb(self, pilots, state=None):

        if self._terminate.is_set():
            return False

        # we register this callback for pilots added to this umgr.  It will
        # specifically look out for pilots which complete, and will make sure
        # that all units are pulled back into umgr control if that happens
        # prematurely.
        #
        # If we find units which have not completed the agent part of the unit
        # state model, we declare them FAILED.  If they can be restarted, we
        # resubmit an identical unit, which then will get a new unit ID.  This
        # avoids state model confusion (the state model is right now expected to
        # be linear), but is not intuitive for the application (FIXME).
        #
        # FIXME: there is a race with the umgr scheduler which may, just now,
        #        and before being notified about the pilot's demise, send new
        #        units to the pilot.

        # we only look into pilot states when the umgr is still active
        # FIXME: note that there is a race in that the umgr can be closed while
        #        we are in the cb.
        # FIXME: `self._closed` is not an `mt.Event`!
        if self._closed:
            self._log.debug('umgr closed, ignore pilot cb %s',
                            ['%s:%s' % (p.uid, p.state) for p in pilots])
            return True

        if not isinstance(pilots, list):
            pilots = [pilots]

        for pilot in pilots:

            state = pilot.state

            if state in rps.FINAL:

                self._log.debug('pilot %s is final - pull units', pilot.uid)

                unit_cursor = self.session._dbs._c.find({
                    'type'    : 'unit',
                    'pilot'   : pilot.uid,
                    'umgr'    : self.uid,
                    'control' : {'$in' : ['agent_pending', 'agent']}})

                if not unit_cursor.count():
                    units = list()
                else:
                    units = list(unit_cursor)

                self._log.debug("units pulled: %3d (pilot dead)", len(units))

                if not units:
                    continue

                # update the units to avoid pulling them again next time.
                # NOTE:  this needs not locking with the unit pulling in the
                #        _unit_pull_cb, as that will only pull umgr_pending
                #        units.
                uids = [unit['uid'] for unit in units]

                self._session._dbs._c.update({'type'  : 'unit',
                                              'uid'   : {'$in'     : uids}},
                                             {'$set'  : {'control' : 'umgr'}},
                                             multi=True)
                to_restart = list()
                for unit in units:

                    unit['state'] = rps.FAILED
                    if not unit['description'].get('restartable'):
                        self._log.debug('unit %s not restartable', unit['uid'])
                        continue

                    self._log.debug('unit %s is  restartable', unit['uid'])
                    unit['restarted'] = True
                    ud = rpcud.ComputeUnitDescription(unit['description'])
                    to_restart.append(ud)
                    # FIXME: increment some restart counter in the description?
                    # FIXME: reference the resulting new uid in the old unit.

                if to_restart and not self._closed:
                    self._log.debug('restart %s units', len(to_restart))
                    restarted = self.submit_units(to_restart)
                    for u in restarted:
                        self._log.debug('restart unit %s', u.uid)

                # final units are not pushed
                self.advance(units, publish=True, push=False)


        # keep cb registered
        return True


    # --------------------------------------------------------------------------
    #
    def _state_pull_cb(self):

        if self._terminate.is_set():
            return False

        # pull all unit states from the DB, and compare to the states we know
        # about.  If any state changed, update the unit instance and issue
        # notification callbacks as needed.  Do not advance the state (again).
        # FIXME: we also pull for dead units.  That is not efficient...
        # FIXME: this needs to be converted into a tailed cursor in the update
        #        worker
        units = self._session._dbs.get_units(umgr_uid=self.uid)

        for unit in units:
            if not self._update_unit(unit, publish=True, advance=False):
                return False

        return True


    # --------------------------------------------------------------------------
    #
    def _unit_pull_cb(self):

        if self._terminate.is_set():
            return False

        # pull units from the agent which are about to get back
        # under umgr control, and push them into the respective queues
        # FIXME: this should also be based on a tailed cursor
        # FIXME: Unfortunately, 'find_and_modify' is not bulkable, so we have
        #        to use 'find'.  To avoid finding the same units over and over
        #        again, we update the 'control' field *before* running the next
        #        find -- so we do it right here.
        unit_cursor = self.session._dbs._c.find({'type'    : 'unit',
                                                 'umgr'    : self.uid,
                                                 'control' : 'umgr_pending'})

        if not unit_cursor.count():
            # no units whatsoever...
          # self._log.info("units pulled:    0")
            return True  # this is not an error

        # update the units to avoid pulling them again next time.
        units = list(unit_cursor)
        uids  = [unit['uid'] for unit in units]

        self._log.info("units pulled:    %d", len(uids))

        for unit in units:
            unit['control'] = 'umgr'

        self._session._dbs._c.update({'type'  : 'unit',
                                      'uid'   : {'$in'     : uids}},
                                     {'$set'  : {'control' : 'umgr'}},
                                     multi=True)

        self._log.info("units pulled: %4d", len(units))
        self._prof.prof('get', msg="bulk size: %d" % len(units), uid=self.uid)
        for unit in units:

            # we need to make sure to have the correct state:
            uid = unit['uid']
            self._prof.prof('get', uid=uid)

            old = unit['state']
            new = rps._unit_state_collapse(unit['states'])

            if old != new:
                self._log.debug("unit  pulled %s: %s / %s", uid, old, new)

            unit['state'] = new

        # now we really own the CUs, and can start working on them (ie. push
        # them into the pipeline).

        to_stage    = list()
        to_finalize = list()

        for unit in units:
            # only advance units to data stager if we need data staging
            # = otherwise finalize them right away
            if unit['description'].get('output_staging'):
                to_stage.append(unit)
            else:
                to_finalize.append(unit)

        # don't profile state transitions - those happened in the past
        if to_stage:
            if self._has_sout:
                # normal route: needs data stager
                self.advance(to_stage, publish=True, push=True, prof=False)
            else:
                self._log.error('output staging needed but not available!')
                for unit in to_stage:
                    unit['target_state'] = rps.FAILED
                    to_finalize.append(unit)

        if to_finalize:
            # shortcut, skip the data stager, but fake state transition
            self.advance(to_finalize, state=rps.UMGR_STAGING_OUTPUT,
                                      publish=True, push=False)

            # move to final stata
            for unit in to_finalize:
                unit['state'] = unit['target_state']
            self.advance(to_finalize, publish=True, push=False)

        return True


    # --------------------------------------------------------------------------
    #
    def _state_sub_cb(self, topic, msg):

        if self._terminate.is_set():
            return False

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd != 'update':
            self._log.debug('ignore state cb msg with cmd %s', cmd)
            return True

        if isinstance(arg, list): things =  arg
        else                    : things = [arg]

        cb_requests = list()

        for thing in things:

            if thing.get('type') == 'unit':

                # we got the state update from the state callback - don't
                # publish it again
                to_notify = self._update_unit(thing, publish=False,
                                              advance=False)
                if to_notify:
                    cb_requests += to_notify
            else:
                self._log.debug('umgr state cb ignores %s/%s', thing.get('uid'),
                        thing.get('state'))

        if cb_requests:
            if _USE_BULK_CB:
                self._bulk_cbs(set([unit for unit,state in cb_requests]))
            else:
                for unit,state in cb_requests:
                    self._unit_cb(unit, state)

        return True


    # --------------------------------------------------------------------------
    #
    def _update_unit(self, unit_dict, publish=False, advance=False):

        uid = unit_dict['uid']

        # return information about needed callback and advance activities, so
        # that we don't break bulks here.
        # note however that individual unit callbacks are still being called on
        # each unit (if any are registered), which can lead to arbitrary,
        # application defined delays.
        to_notify = list()

        with self._units_lock:

            # we don't care about units we don't know
            if uid not in self._units:
                self._log.debug('umgr: unknown: %s', uid)
                return None

            unit = self._units[uid]

            # only update on state changes
            current = unit.state
            target  = unit_dict['state']
            if current == target:
                self._log.debug('umgr: static: %s', uid)
                return None

            target, passed = rps._unit_state_progress(uid, current, target)

            if target in [rps.CANCELED, rps.FAILED]:
                # don't replay intermediate states
                passed = passed[-1:]

            for s in passed:
                unit_dict['state'] = s
                self._units[uid]._update(unit_dict)
                to_notify.append([unit, s])

                # we don't usually advance state at this point, but just keep up
                # with state changes reported from elsewhere
                if advance:
                    self.advance(unit_dict, s, publish=publish, push=False,
                                 prof=False)

            self._log.debug('umgr: notify: %s %s %s', len(to_notify), unit_dict,
                    unit_dict['state'])
            return to_notify


    # --------------------------------------------------------------------------
    #
    def _unit_cb(self, unit, state):

        with self._cb_lock:

            uid      = unit.uid
            cb_dicts = list()
            metric   = rpc.UNIT_STATE

            # get wildcard callbacks
            cb_dicts += self._callbacks[metric].get('*', {}).values()
            cb_dicts += self._callbacks[metric].get(uid, {}).values()

            for cb_dict in cb_dicts:

                cb           = cb_dict['cb']
                cb_data      = cb_dict['cb_data']

                try:
                    if cb_data: cb(unit, state, cb_data)
                    else      : cb(unit, state)
                except:
                    self._log.exception('cb error (%s)', cb.__name__)


    # --------------------------------------------------------------------------
    #
    def _bulk_cbs(self, units,  metrics=None):

        if not metrics: metrics = [rpc.UNIT_STATE]
        else          : metrics = ru.as_list(metrics)

        cbs = dict()  # bulked callbacks to call

        with self._cb_lock:

            for metric in metrics:

                # get wildcard callbacks
                cb_dicts = self._callbacks[metric].get('*')
                for cb_name in cb_dicts:
                    cbs[cb_name] = {'cb'     : cb_dicts[cb_name]['cb'],
                                    'cb_data': cb_dicts[cb_name]['cb_data'],
                                    'units'  : set(units)}

                # add unit specific callbacks if needed
                for unit in units:

                    uid = unit.uid
                    if uid not in self._callbacks[metric]:
                        continue

                    cb_dicts = self._callbacks[metric].get(uid, {})
                    for cb_name in cb_dicts:

                        if cb_name in cbs:
                            cbs[cb_name]['units'].add(unit)
                        else:
                            cbs[cb_name] = {'cb'     : cb_dicts[cb_name]['cb'],
                                            'cb_data': cb_dicts[cb_name]['cb_data'],
                                            'units'  : set([unit])}

            for cb_name in cbs:

                cb      = cbs[cb_name]['cb']
                cb_data = cbs[cb_name]['cb_data']
                objs    = cbs[cb_name]['units']

                if cb_data: cb(list(objs), cb_data)
                else      : cb(list(objs))


    # --------------------------------------------------------------------------
    #
    # FIXME: this needs to go to the scheduler
    def _default_wait_queue_size_cb(self, umgr, wait_queue_size):

        # FIXME: this needs to come from the scheduler?
        if self._terminate.is_set():
            return False

        self._log.info("[Callback]: wait_queue_size: %s.", wait_queue_size)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        """
        Returns the unique id.
        """
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def scheduler(self):
        """
        Returns the scheduler name.
        """

        return self._cfg.get('scheduler')



    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pilots):
        """
        Associates one or more pilots with the unit manager.

        **Arguments:**

            * **pilots** [:class:`radical.pilot.ComputePilot` or list of
              :class:`radical.pilot.ComputePilot`]: The pilot objects that will be
              added to the unit manager.
        """

        if not isinstance(pilots, list):
            pilots = [pilots]

        if len(pilots) == 0:
            raise ValueError('cannot add no pilots')

        with self._pilots_lock:

            # sanity check, and keep pilots around for inspection
            for pilot in pilots:
                pid = pilot.uid
                if pid in self._pilots:
                    raise ValueError('pilot %s already added' % pid)
                self._pilots[pid] = pilot

                # subscribe for state updates
                pilot.register_callback(self._pilot_state_cb)

        pilot_docs = [pilot.as_dict() for pilot in pilots]

        # publish to the command channel for the scheduler to pick up
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'add_pilots',
                                          'arg' : {'pilots': pilot_docs,
                                                   'umgr'  : self.uid}})


    # --------------------------------------------------------------------------
    #
    def list_pilots(self):
        """
        Lists the UIDs of the pilots currently associated with the unit manager.

        **Returns:**
              * A list of :class:`radical.pilot.ComputePilot` UIDs [`string`].
        """

        with self._pilots_lock:
            return list(self._pilots.keys())


    # --------------------------------------------------------------------------
    #
    def get_pilots(self):
        """
        Get the pilots instances currently associated with the unit manager.

        **Returns:**
              * A list of :class:`radical.pilot.ComputePilot` instances.
        """

        with self._pilots_lock:
            return list(self._pilots.values())


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pilot_ids, drain=False):
        """
        Disassociates one or more pilots from the unit manager.

        After a pilot has been removed from a unit manager, it won't process
        any of the unit manager's units anymore. Calling `remove_pilots`
        doesn't stop the pilot itself.

        **Arguments:**

            * **drain** [`boolean`]: Drain determines what happens to the units
              which are managed by the removed pilot(s). If `True`, all units
              currently assigned to the pilot are allowed to finish execution.
              If `False` (the default), then non-final units will be canceled.
        """

        # TODO: Implement 'drain'.
        # NOTE: the actual removal of pilots from the scheduler is asynchron!

        if drain:
            raise RuntimeError("'drain' is not yet implemented")

        if not isinstance(pilot_ids, list):
            pilot_ids = [pilot_ids]

        if len(pilot_ids) == 0:
            raise ValueError('cannot remove no pilots')

        with self._pilots_lock:

            # sanity check, and keep pilots around for inspection
            for pid in pilot_ids:
                if pid not in self._pilots:
                    raise ValueError('pilot %s not removed' % pid)
                del(self._pilots[pid])

        # publish to the command channel for the scheduler to pick up
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'remove_pilots',
                                          'arg' : {'pids'  : pilot_ids,
                                                   'umgr'  : self.uid}})


    # --------------------------------------------------------------------------
    #
    def list_units(self):
        """
        Returns the UIDs of the :class:`radical.pilot.ComputeUnit` managed by
        this unit manager.

        **Returns:**
              * A list of :class:`radical.pilot.ComputeUnit` UIDs [`string`].
        """

        with self._pilots_lock:
            return list(self._units.keys())


    # --------------------------------------------------------------------------
    #
    def submit_units(self, descriptions):
        """
        Submits on or more :class:`radical.pilot.ComputeUnit` instances to the
        unit manager.

        **Arguments:**
            * **descriptions** [:class:`radical.pilot.ComputeUnitDescription`
              or list of :class:`radical.pilot.ComputeUnitDescription`]: The
              description of the compute unit instance(s) to create.

        **Returns:**
              * A list of :class:`radical.pilot.ComputeUnit` objects.
        """

        from .compute_unit import ComputeUnit

        ret_list = True
        if not isinstance(descriptions, list):
            ret_list     = False
            descriptions = [descriptions]

        if len(descriptions) == 0:
            raise ValueError('cannot submit no unit descriptions')

        # we return a list of compute units
        self._rep.progress_tgt(len(descriptions), label='submit')
        units = list()
        for ud in descriptions:

            if not ud.executable:
                raise ValueError('compute unit executable must be defined')

            unit = ComputeUnit(umgr=self, descr=ud)
            units.append(unit)

            # keep units around
            with self._units_lock:
                self._units[unit.uid] = unit

            if self._session._rec:
                ru.write_json(ud.as_dict(), "%s/%s.batch.%03d.json"
                        % (self._session._rec, unit.uid, self._rec_id))

            self._rep.progress()

        self._rep.progress_done()

        if self._session._rec:
            self._rec_id += 1

        # insert units into the database, as a bulk.
        unit_docs = [u.as_dict() for u in units]
        self._session._dbs.insert_units(unit_docs)

        # Only after the insert can we hand the units over to the next
        # components (ie. advance state).
        self.advance(unit_docs, rps.UMGR_SCHEDULING_PENDING,
                     publish=True, push=True)

        if ret_list: return units
        else       : return units[0]


    # --------------------------------------------------------------------------
    #
    def get_units(self, uids=None):
        """Returns one or more compute units identified by their IDs.

        **Arguments:**
            * **uids** [`string` or `list of strings`]: The IDs of the
              compute unit objects to return.

        **Returns:**
              * A list of :class:`radical.pilot.ComputeUnit` objects.
        """

        if not uids:
            with self._units_lock:
                ret = list(self._units.values())
            return ret

        ret_list = True
        if (not isinstance(uids, list)) and (uids is not None):
            ret_list = False
            uids = [uids]

        ret = list()
        with self._units_lock:
            for uid in uids:
                if uid not in self._units:
                    raise ValueError('unit %s not known' % uid)
                ret.append(self._units[uid])

        if ret_list: return ret
        else       : return ret[0]


    # --------------------------------------------------------------------------
    #
    def wait_units(self, uids=None, state=None, timeout=None):
        """
        Returns when one or more :class:`radical.pilot.ComputeUnits` reach a
        specific state.

        If `uids` is `None`, `wait_units` returns when **all**
        ComputeUnits reach the state defined in `state`.  This may include
        units which have previously terminated or waited upon.

        **Example**::

            # TODO -- add example

        **Arguments:**

            * **uids** [`string` or `list of strings`]
              If uids is set, only the ComputeUnits with the specified
              uids are considered. If uids is `None` (default), all
              ComputeUnits are considered.

            * **state** [`string`]
              The state that ComputeUnits have to reach in order for the call
              to return.

              By default `wait_units` waits for the ComputeUnits to
              reach a terminal state, which can be one of the following:

              * :data:`radical.pilot.rps.DONE`
              * :data:`radical.pilot.rps.FAILED`
              * :data:`radical.pilot.rps.CANCELED`

            * **timeout** [`float`]
              Timeout in seconds before the call returns regardless of Pilot
              state changes. The default value **None** waits forever.
        """

        if not uids:
            with self._units_lock:
                uids = list()
                for uid,unit in self._units.items():
                    if unit.state not in rps.FINAL:
                        uids.append(uid)

        if   not state                  : states = rps.FINAL
        elif not isinstance(state, list): states = [state]
        else                            : states =  state

        # we simplify state check by waiting for the *earliest* of the given
        # states - if the unit happens to be in any later state, we are sure the
        # earliest has passed as well.
        check_state_val = rps._unit_state_values[rps.FINAL[-1]]
        for state in states:
            check_state_val = min(check_state_val,
                                  rps._unit_state_values[state])

        ret_list = True
        if not isinstance(uids, list):
            ret_list = False
            uids = [uids]

        start    = time.time()
        to_check = None

        with self._units_lock:
            to_check = [self._units[uid] for uid in uids]

        # We don't want to iterate over all units again and again, as that would
        # duplicate checks on units which were found in matching states.  So we
        # create a list from which we drop the units as we find them in
        # a matching state
        self._rep.progress_tgt(len(to_check), label='wait')
        while to_check and not self._terminate.is_set():

            # check timeout
            if timeout and (timeout <= (time.time() - start)):
                self._log.debug ("wait timed out")
                break

            time.sleep (0.1)

            # FIXME: print percentage...
          # print 'wait units: %s' % [[u.uid, u.state] for u in to_check]

            check_again = list()
            for unit in to_check:

                # we actually don't check if a unit is in a specific (set of)
                # state(s), but rather check if it ever *has been* in any of
                # those states
                if unit.state not in rps.FINAL and \
                    rps._unit_state_values[unit.state] < check_state_val:
                    # this unit does not match the wait criteria
                    check_again.append(unit)

                else:
                    # stop watching this unit
                    if unit.state in [rps.FAILED]:
                        self._rep.progress()  # (color='error', c='-')
                    elif unit.state in [rps.CANCELED]:
                        self._rep.progress()  # (color='warn', c='*')
                    else:
                        self._rep.progress()  # (color='ok', c='+')

            to_check = check_again

        self._rep.progress_done()


        # grab the current states to return
        state = None
        with self._units_lock:
            states = [self._units[uid].state for uid in uids]

        sdict = {state: states.count(state) for state in set(states)}
        for state in sorted(set(states)):
            self._rep.info('\t%-10s: %5d\n' % (state, sdict[state]))

        if to_check: self._rep.warn('>>timeout\n')
        else       : self._rep.ok  ('>>ok\n')

        # done waiting
        if ret_list: return states
        else       : return states[0]


    # --------------------------------------------------------------------------
    #
    def cancel_units(self, uids=None):
        """
        Cancel one or more :class:`radical.pilot.ComputeUnits`.

        Note that cancellation of units is *immediate*, i.e. their state is
        immediately set to `CANCELED`, even if some RP component may still
        operate on the units.  Specifically, other state transitions, including
        other final states (`DONE`, `FAILED`) can occur *after* cancellation.
        This is a side effect of an optimization: we consider this
        acceptable tradeoff in the sense "Oh, that unit was DONE at point of
        cancellation -- ok, we can use the results, sure!".

        If that behavior is not wanted, set the environment variable:

            export RADICAL_PILOT_STRICT_CANCEL=True

        **Arguments:**
            * **uids** [`string` or `list of strings`]: The IDs of the
              compute units objects to cancel.
        """

        if not uids:
            with self._units_lock:
                uids  = list(self._units.keys())
        else:
            if not isinstance(uids, list):
                uids = [uids]

        # NOTE: We advance all units to cancelled, and send a cancellation
        #       control command.  If that command is picked up *after* some
        #       state progression, we'll see state transitions after cancel.
        #       For non-final states that is not a problem, as it is equivalent
        #       with a state update message race, which our state collapse
        #       mechanism accounts for.  For an eventual non-canceled final
        #       state, we do get an invalid state transition.  That is also
        #       corrected eventually in the state collapse, but the point
        #       remains, that the state model is temporarily violated.  We
        #       consider this a side effect of the fast-cancel optimization.
        #
        #       The env variable 'RADICAL_PILOT_STRICT_CANCEL == True' will
        #       disable this optimization.
        #
        # FIXME: the effect of the env var is not well tested
        if 'RADICAL_PILOT_STRICT_CANCEL' not in os.environ:
            with self._units_lock:
                units = [self._units[uid] for uid  in uids ]
            unit_docs = [unit.as_dict()   for unit in units]
            self.advance(unit_docs, state=rps.CANCELED, publish=True, push=True)

        # we *always* issue the cancellation command to the local components
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_units',
                                          'arg' : {'uids' : uids,
                                                   'umgr' : self.uid}})

        # we also inform all pilots about the cancelation request
        self._session._dbs.pilot_command(cmd='cancel_units', arg={'uids':uids})

        # In the default case of calling 'advance' above, we just set the state,
        # so we *know* units are canceled.  But we nevertheless wait until that
        # state progression trickled through, so that the application will see
        # the same state on unit inspection.
        self.wait_units(uids=uids)


    # --------------------------------------------------------------------------
    #
    def register_callback(self, cb, cb_data=None, metric=None, uid=None):
        """
        Registers a new callback function with the UnitManager.  Manager-level
        callbacks get called if the specified metric changes.  The default
        metric `UNIT_STATE` fires the callback if any of the ComputeUnits
        managed by the PilotManager change their state.

        All callback functions need to have the same signature::

            def cb(obj, value)

        where ``object`` is a handle to the object that triggered the callback,
        ``value`` is the metric, and ``data`` is the data provided on
        callback registration..  In the example of `UNIT_STATE` above, the
        object would be the unit in question, and the value would be the new
        state of the unit.

        If 'cb_data' is given, then the 'cb' signature changes to

            def cb(obj, state, cb_data)

        and 'cb_data' are passed unchanged.

        If 'uid' is given, the callback will invoked only for the specified
        unit.


        Available metrics are:

          * `UNIT_STATE`: fires when the state of any of the units which are
            managed by this unit manager instance is changing.  It communicates
            the unit object instance and the units new state.

          * `WAIT_QUEUE_SIZE`: fires when the number of unscheduled units (i.e.
            of units which have not been assigned to a pilot for execution)
            changes.
        """

        # FIXME: the signature should be (self, metrics, cb, cb_data)

        if not metric:
            metric = rpc.UNIT_STATE

        if  metric not in rpc.UMGR_METRICS:
            raise ValueError ("Metric '%s' not available on the umgr" % metric)

        if not uid:
            uid = '*'

        elif uid not in self._units:
            raise ValueError('no such unit %s' % uid)


        with self._cb_lock:
            cb_name = cb.__name__

            if metric not in self._callbacks:
                self._callbacks[metric] = dict()

            if uid not in self._callbacks[metric]:
                self._callbacks[metric][uid] = dict()

            self._callbacks[metric][uid][cb_name] = {'cb'      : cb,
                                                     'cb_data' : cb_data}


    # --------------------------------------------------------------------------
    #
    def unregister_callback(self, cb=None, metrics=None, uid=None):

        if not metrics: metrics = [rpc.UMGR_METRICS]
        else          : metrics = ru.as_list(metrics)

        if not uid:
            uid = '*'

        elif uid not in self._units:
            raise ValueError('no such unit %s' % uid)

        for metric in metrics:
            if metric not in rpc.UMGR_METRICS :
                raise ValueError ("invalid umgr metric '%s'" % metric)

        with self._cb_lock:

            for metric in metrics:

                if metric not in rpc.UMGR_METRICS :
                    raise ValueError("cb metric '%s' unknown" % metric)

                if metric not in self._callbacks:
                    raise ValueError("cb metric '%s' invalid" % metric)

                if uid not in self._callbacks[metric]:
                    raise ValueError("cb target '%s' invalid" % uid)

                if cb:
                    to_delete = [cb.__name__]
                else:
                    to_delete = list(self._callbacks[metric][uid].keys())

                for cb_name in to_delete:

                    if cb_name not in self._callbacks[uid][metric]:
                        raise ValueError("cb %s not registered" % cb_name)

                    del(self._callbacks[uid][metric][cb_name])


# ------------------------------------------------------------------------------




#!/usr/bin/env python 

import os
import sys

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def log(msg):

    sys.stdout.write('%s\n' % msg)
    sys.stdout.flush()


# ------------------------------------------------------------------------------
#
# WORKER
#
if __name__ == '__main__':

    uid          = os.environ['RP_UNIT_ID']
    addr_work    = os.environ['RP_WORK_QUEUE_OUT']
    addr_res     = os.environ['RP_RESULT_QUEUE_IN']

    work_queue   = ru.Getter('RP_WORK_QUEUE',   addr_work)
    result_queue = ru.Putter('RP_RESULT_QUEUE', addr_res)

    work = work_queue.get()
    while work:

        result = '%s: %s' % (uid, work)
        log('= %s' % result)
        result_queue.put(result)
        work = work_queue.get()

        if work == ['QUIT']:
            break

    log('w DONE %s' % uid)


# ------------------------------------------------------------------------------

